# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0
 
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
from collections import defaultdict
import glob
import logging
import os
import re
from naarad.graphing.plot_data import PlotData as PD
import naarad.utils
import naarad.httpdownload
import naarad.naarad_constants as CONSTANTS

logger = logging.getLogger('naarad.metrics.metric')

class Metric(object):

  def __init__(self, metric_type, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end,
                rule_strings, important_sub_metrics, **other_options):
    self.metric_type = metric_type
    self.infile_list = infile_list
    self.hostname = hostname
    self.outdir = output_directory
    self.resource_path = resource_path
    self.resource_directory = os.path.join(self.outdir, self.resource_path)
    self.label = label
    self.ts_start = naarad.utils.get_standardized_timestamp(ts_start, None)
    self.ts_end = naarad.utils.get_standardized_timestamp(ts_end, None)
    self.calc_metrics = None
    self.precision = None
    self.sep = ','
    self.titles_string = None
    self.ylabels_string = None
    self.csv_files = []
    self.plot_files = []
    self.stats_files = []
    self.important_stats_files = []
    self.percentiles_files = []
    self.column_csv_map = {}
    self.csv_column_map = {}
    self.sub_metric_description = defaultdict(lambda: 'None')  # the description of the submetrics. 
    self.sub_metric_unit = defaultdict(lambda: 'None')      # the unit of the submetrics.  The plot will have the Y-axis being: Metric name (Unit)
    self.important_sub_metrics = important_sub_metrics
    self.sla_list = []  # TODO : remove this once report has grading done in the metric tables
    self.sla_map = defaultdict(lambda :defaultdict(lambda: defaultdict(None)))
    self.calculated_stats = {}
    self.calculated_percentiles = {}
    self.summary_stats_list = CONSTANTS.DEFAULT_SUMMARY_STATS
    self.summary_stats = defaultdict(dict)
    self.status = CONSTANTS.OK
    self.ignore = False
    self.timezone = "PDT"
    self.options = None
    self.sub_metrics = None   #users can specify what sub_metrics to process/plot;
    self.groupby = None
    self.summary_charts = []
    # Leave the flag here for the future use to control summary page
    self.summary_html_content_enabled = True
    for (key, val) in rule_strings.iteritems():
      naarad.utils.set_sla(self, self.label, key, val)
    if other_options:
      for (key, val) in other_options.iteritems():
        setattr(self, key, val)
      if not self.titles_string:
        self.titles_string = self.columns
      if self.columns:
        self.columns = self.columns.split()
      if self.groupby:
        self.groupby = self.groupby.split()

      self.titles = dict(zip(self.columns, self.titles_string.split(','))) if self.columns and self.titles_string else None
      self.ylabels = dict(zip(self.columns, self.ylabels_string.split(','))) if self.columns and self.ylabels_string else None

  def name_to_index(self, name):
    index = None
    for i in range(len(self.columns)):
      if name == self.columns[i]:
        index = i+1
    return index

  def get_groupby_indexes(self, groupby):
    groupby_indexes = []
    for group in groupby:
      if ':' in group.rstrip(':'):
        name, index = group.split(':')
        groupby_indexes.append(index)
      else:
        groupby_indexes.append(self.name_to_index(group.rstrip(':')))
    return groupby_indexes

  def ts_out_of_range(self, timestamp):
    if self.ts_start and timestamp < self.ts_start:
      return True
    elif self.ts_end and timestamp > self.ts_end:
      return True
    return False

  def collect_local(self, infile):
    return os.path.exists(infile)

  def collect(self):
    # self.infile_list can be of several formats: for instance a local dir (e.g., /path/a.log) or an http url;
    # decide the case based on self.infile_list;
    # self.access is optional, can be removed.
    collected_files = []
    for infile in self.infile_list:
      if infile.startswith("http://") or infile.startswith("https://"):
        if naarad.utils.is_valid_url(infile):
          http_download_dir = os.path.join(self.outdir, self.label)
          output_file = naarad.httpdownload.download_url_single(infile, http_download_dir)
          if output_file:
            collected_files.append(output_file)
          else:
            return False
        else:
          logger.error("The given url of {0} is invalid.\n".format(infile))
          return False
      else:
        file_matches = glob.glob(infile)
        if len(file_matches) == 0:
          return False
        for file_name in file_matches:
          if self.collect_local(file_name):
            collected_files.append(file_name)
          else:
            return False
    self.infile_list = collected_files
    return True

  def get_csv(self, column, groupby=None):
    if groupby:
      column = groupby + '.' + column
    if column in self.column_csv_map.keys():
      return self.column_csv_map[column]
    col = naarad.utils.sanitize_string(column)
    csv = os.path.join(self.resource_directory, self.label + '.' + col + '.csv')
    self.csv_column_map[csv] = column
    self.column_csv_map[column] = csv
    return csv
  
  def get_important_sub_metrics_csv(self):
    csv = os.path.join(self.resource_directory, self.label + '.important_sub_metrics.csv')
    return csv

  def get_stats_csv(self):
    csv = os.path.join(self.resource_directory, self.label + '.stats.csv')
    return csv

  def get_percentiles_csv_from_data_csv(self, data_csv):
    percentile_csv_file = '.'.join(data_csv.split('.')[0:-1]) + '.percentiles.csv'
    return percentile_csv_file

  def get_sla_csv(self):
    csv = os.path.join(self.resource_directory, self.label + '.sla.csv')
    return csv

  def update_summary_stats(self, column):
    for stat in self.summary_stats_list:
      if stat.startswith('p'):
        self.summary_stats[column][stat] = naarad.utils.normalize_float_for_display(self.calculated_percentiles[column][int(stat[1:])])
      else:
        self.summary_stats[column][stat] = naarad.utils.normalize_float_for_display(self.calculated_stats[column][stat])

  def parse(self):
    qps = defaultdict(int)
    groupby_idxes = None
    if self.groupby:
      groupby_idxes = self.get_groupby_indexes(self.groupby)
    data = defaultdict(list)
    aggregate_data = defaultdict(list)
    for input_file in self.infile_list:
      logger.info("Working on " + input_file)
      timestamp_format = None
      with open(input_file, 'r') as infile:
        for line in infile:
          if self.sep is None or self.sep == '':
            words = line.strip().split()
          else:
            words = line.strip().split(self.sep)
          if len(words) == 0:
            continue
          if len(words) <= len(self.columns): #NOTE: len(self.columns) is always one less than len(words) since we assume the very first column is timestamp
            logger.warning("WARNING: Number of columns given in config is more than number of columns present in line {0}\n", line)
            continue
          if not timestamp_format or timestamp_format == 'unknown':
            timestamp_format = naarad.utils.detect_timestamp_format(words[0])
          if timestamp_format == 'unknown':
            continue
          ts = naarad.utils.get_standardized_timestamp(words[0], timestamp_format)
          if ts == -1:
            continue
          ts = naarad.utils.reconcile_timezones(ts, self.timezone, self.graph_timezone)
          if self.ts_out_of_range(ts):
            continue
          qps[ts.split('.')[0]] += 1
          if self.groupby:
            groupby_names = None
            for index in groupby_idxes:
              if not groupby_names:
                groupby_names = words[index].rstrip(':')
              else:
                groupby_names += '.' + words[index].rstrip(':')
            for i in range(len(self.columns)):
              if i+1 in groupby_idxes:
                continue
              else:
                out_csv = self.get_csv(self.columns[i], groupby_names)
                data[out_csv].append(ts + ',' + words[i+1])
                out_csv = self.get_csv(self.columns[i], 'Overall_Summary')
                aggregate_data[out_csv].append((ts, words[i+1]))
          else:
            for i in range(len(self.columns)):
              out_csv = self.get_csv(self.columns[i])
              if out_csv in data:
                data[out_csv].append(ts + ',' + words[i+1])
              else:
                data[out_csv] = []
                data[out_csv].append(ts + ',' + words[i+1])
    # Post processing, putting data in csv files
    data[self.get_csv('qps')] = map(lambda x: x[0] + ',' + str(x[1]), sorted(qps.items()))
    for csv in data.keys():
      self.csv_files.append(csv)
      with open(csv, 'w') as fh:
        fh.write('\n'.join(sorted(data[csv])))
    if self.groupby:
      for csv in aggregate_data.keys():
        new_data = defaultdict(float)
        for timestamp, value in aggregate_data[csv]:
          new_data[timestamp] += float(value)
        aggregate_data[csv] = []
        for ts, value in sorted(new_data.items()):
          aggregate_data[csv].append(str(ts) + ',' + str(value))
        self.csv_files.append(csv)
        with open(csv, 'w') as fh:
          fh.write('\n'.join(sorted(aggregate_data[csv])))
    return True

  def calculate_stats(self):
    stats_to_calculate = ['mean', 'std', 'min', 'max']  # TODO: get input from user
    percentiles_to_calculate = range(0, 100, 1)  # TODO: get input from user
    headers = CONSTANTS.SUBMETRIC_HEADER + ',mean,std,p50,p75,p90,p95,p99,min,max\n'  # TODO: This will be built from user input later on
    metric_stats_csv_file = self.get_stats_csv()
    imp_metric_stats_csv_file = self.get_important_sub_metrics_csv()
    imp_metric_stats_present = False  
    metric_stats_present = False
    logger.info("Calculating stats for important sub-metrics in %s and all sub-metrics in %s", imp_metric_stats_csv_file, metric_stats_csv_file)
    with open(metric_stats_csv_file, 'w') as FH_W:
      with open(imp_metric_stats_csv_file, 'w') as FH_W_IMP:
        for csv_file in self.csv_files:
          data = []
          value_error = False
          if not os.path.getsize(csv_file):
            continue
          column = self.csv_column_map[csv_file]
          percentile_csv_file = self.get_percentiles_csv_from_data_csv(csv_file)
          with open(csv_file, 'r') as FH:
            for line in FH:
              words = line.split(',')
              try:
                data.append(float(words[1]))
              except ValueError:
                if not value_error:
                  logger.error("Cannot convert to float. Some data is ignored in file " + csv_file)
                  value_error = True
                continue
          self.calculated_stats[column], self.calculated_percentiles[column] = naarad.utils.calculate_stats(data, stats_to_calculate, percentiles_to_calculate)
          
          with open(percentile_csv_file, 'w') as FH_P:
            for percentile in sorted(self.calculated_percentiles[column].iterkeys()):
              FH_P.write("%d, %f\n" % (percentile, self.calculated_percentiles[column][percentile]))
          self.percentiles_files.append(percentile_csv_file)
          self.update_summary_stats(column)
          to_write = [column, self.calculated_stats[column]['mean'], self.calculated_stats[column]['std'], self.calculated_percentiles[column][50], self.calculated_percentiles[column][75], self.calculated_percentiles[column][90], self.calculated_percentiles[column][95], self.calculated_percentiles[column][99], self.calculated_stats[column]['min'], self.calculated_stats[column]['max']]
          to_write = map(lambda x: naarad.utils.normalize_float_for_display(x), to_write)
          if not metric_stats_present:
            metric_stats_present = True
            FH_W.write(headers)
          FH_W.write(','.join(to_write) + '\n') 
          # Important sub-metrics and their stats go in imp_metric_stats_csv_file
          sub_metric = column
          if self.metric_type in self.device_types:
            sub_metric = column.split('.')[1]
          if self.check_important_sub_metrics(sub_metric):
            if not imp_metric_stats_present:
              FH_W_IMP.write(headers)
              imp_metric_stats_present = True
            FH_W_IMP.write(','.join(to_write) + '\n')
        if imp_metric_stats_present:
          self.important_stats_files.append(imp_metric_stats_csv_file)
      self.stats_files.append(metric_stats_csv_file)

  def calc(self):
    if not self.calc_metrics:
      return
    calculation_array = self.calc_metrics.split()
    for calculation in calculation_array:
      words = calculation.split('=')
      newmetric = words[0]
      expr = words[1]
      p = re.compile('(\w+)\((.+)\)')
      calc_type = p.match(expr).group(1)
      old_metric = p.match(expr).group(2)
      logger.debug('In calc() : %s %s %s %s', newmetric, expr, old_metric, calc_type)
      if not calc_type in ('rate', 'diff'):
        logger.error("ERROR: Invalid calc_metric type {0} defined in config".format(calc_type))
        continue
      old_metric_csv = self.get_csv(old_metric)
      new_metric_csv = self.get_csv(newmetric)
      self.csv_files.append(new_metric_csv)
      old_val = None
      with open(old_metric_csv, 'r') as FH:
        with open(new_metric_csv, 'w') as NEW_FH:
          for line in FH:
            w = line.split(',')
            ts = w[0]
            val = w[1]
            if not old_val:
              old_ts = ts
              old_val = val
              continue
            if calc_type == 'rate':
              #Multiply rate by 1000 since timestamp is in ms
              ts_diff = naarad.utils.convert_to_unixts(ts) - naarad.utils.convert_to_unixts(old_ts)
              if ts_diff != 0 :
                new_metric_val = 1000 * (float(val) - float(old_val)) / ts_diff
              else:
                new_metric_val = 0
                logger.warn("rate calculation encountered zero timestamp difference")
            elif calc_type == 'diff':
              new_metric_val = (float(val) - float(old_val))
            old_ts = ts
            old_val = val
            NEW_FH.write(ts)
            NEW_FH.write(',')
            NEW_FH.write(str(new_metric_val))
            NEW_FH.write('\n')

  def plot_timeseries(self, graphing_library = 'matplotlib'):
    """
    plot timeseries for sub-metrics 
    """
    if self.groupby:
      plot_data = {}
      # plot time series data for submetrics
      for out_csv in sorted(self.csv_files, reverse=True):
        csv_filename = os.path.basename(out_csv)
        # The last element is .csv, don't need that in the name of the chart
        column = csv_filename.split('.')[-2]
        transaction_name = ' '.join(csv_filename.split('.')[1:-2])
        plot = PD(input_csv=out_csv, csv_column=1, series_name=transaction_name + '.' + column, y_label=column + ' (' + self.sub_metric_description[column] + ')', precision=None, graph_height=500, graph_width=1200, graph_type='line')
        if transaction_name in plot_data:
          plot_data[transaction_name].append(plot)
        else:
          plot_data[transaction_name] = [plot]
      for transaction in plot_data:
        graphed, div_file = Metric.graphing_modules[graphing_library].graph_data(plot_data[transaction], self.resource_directory, self.resource_path, self.label + '.' + transaction )
        if graphed:
          self.plot_files.append(div_file)
    else:
      graphed = False
      for out_csv in self.csv_files:
        csv_filename = os.path.basename(out_csv)
        # The last element is .csv, don't need that in the name of the chart
        column = self.csv_column_map[out_csv]
        column = naarad.utils.sanitize_string(column)
        graph_title = '.'.join(csv_filename.split('.')[0:-1])
        if self.sub_metric_description and column in self.sub_metric_description.keys():
          graph_title += ' ('+self.sub_metric_description[column]+')'
        if self.sub_metric_unit and column in self.sub_metric_unit.keys():
          plot_data = [PD(input_csv=out_csv, csv_column=1, series_name=graph_title, y_label=column +' ('+ self.sub_metric_unit[column]+')', precision=None, graph_height=600, graph_width=1200, graph_type='line')]
        else:
          plot_data = [PD(input_csv=out_csv, csv_column=1, series_name=graph_title, y_label=column, precision=None, graph_height=600, graph_width=1200, graph_type='line')]
        graphed, div_file = Metric.graphing_modules[graphing_library].graph_data(plot_data, self.resource_directory, self.resource_path, graph_title)
        if graphed:
          self.plot_files.append(div_file)
    return True
  
  def check_important_sub_metrics(self, sub_metric):
    """
    check whether the given sub metric is in important_sub_metrics list 
    """
    if not self.important_sub_metrics:
      return False
    if sub_metric in self.important_sub_metrics:
      return True
    items = sub_metric.split('.')
    if items[-1] in self.important_sub_metrics:
      return True
    return False

  def plot_cdf(self, graphing_library = 'matplotlib'):
    """
    plot CDF for important sub-metrics 
    """
    graphed = False
    for percentile_csv in self.percentiles_files:
      csv_filename = os.path.basename(percentile_csv)
      # The last element is .csv, don't need that in the name of the chart
      column = self.csv_column_map[percentile_csv.replace(".percentiles.", ".")]
      if not self.check_important_sub_metrics(column):
        continue
      column = naarad.utils.sanitize_string(column)
      graph_title = '.'.join(csv_filename.split('.')[0:-1])
      if self.sub_metric_description and column in self.sub_metric_description.keys():
        graph_title += ' ('+self.sub_metric_description[column]+')'
      if self.sub_metric_unit and column in self.sub_metric_unit.keys():
        plot_data = [PD(input_csv=percentile_csv, csv_column=1, series_name=graph_title, x_label='Percentiles', y_label=column +' ('+ self.sub_metric_unit[column]+')', precision=None, graph_height=600, graph_width=1200, graph_type='line')]
      else:
        plot_data = [PD(input_csv=percentile_csv, csv_column=1, series_name=graph_title, x_label='Percentiles', y_label=column, precision=None, graph_height=600, graph_width=1200, graph_type='line')]
      graphed, div_file = Metric.graphing_modules[graphing_library].graph_data_on_the_same_graph(plot_data, self.resource_directory, self.resource_path, graph_title)
      if graphed:
        self.plot_files.append(div_file)
    return True

  def graph(self, graphing_library = 'matplotlib'):
    """ 
    graph generates two types of graphs
    'time': generate a time-series plot for all submetrics (the x-axis is a time series)
    'cdf': generate a CDF plot for important submetrics (the x-axis shows percentiles)
    """
    logger.info('Using graphing_library {lib} for metric {name}'.format(lib=graphing_library, name=self.label))
    self.plot_cdf(graphing_library)
    self.plot_timeseries(graphing_library)
    return True



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");?you may not use this file except in compliance with the License.?You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software?distributed under the License is distributed on an "AS IS" BASIS,?WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""

import ConfigParser
import os
import sys
import time
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'src')))
from naarad import Naarad
import naarad.naarad_constants as CONSTANTS
naarad_obj = None

def setup_module():
  global naarad_obj
  naarad_obj = Naarad()

def test_naarad_apis():
  """
  :return: None
  """
  examples_directory = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'examples')
  config_file = os.path.join(os.path.join(examples_directory, 'conf'),'config-gc')
  config_object = ConfigParser.ConfigParser()
  config_object.optionxform = str
  config_object.read(config_file)
  input_directory = os.path.join(examples_directory,'logs')
  output_directory = 'test_api_temp' 
  diff_output_directory = 'test_api_temp/diff_location'
  report1_location = 'test_api_temp/0'
  report2_location = 'test_api_temp/1'
  global naarad_obj
  test_id_1 = naarad_obj.signal_start(config_file)
  time.sleep(60)
  naarad_obj.signal_stop(test_id_1)
  test_id_2 = naarad_obj.signal_start(config_object)
  time.sleep(60)
  naarad_obj.signal_stop(test_id_2)
  if naarad_obj.analyze(input_directory, output_directory) != CONSTANTS.OK :
    print naarad_obj.get_failed_analyses()
  naarad_obj.get_sla_data(test_id_1)
  naarad_obj.get_stats_data(test_id_1)
  naarad_obj.get_sla_data(test_id_2)
  naarad_obj.get_stats_data(test_id_2)
  if naarad_obj.diff(test_id_1, test_id_2, None) != CONSTANTS.OK:
    print 'Error encountered during diff'
  if naarad_obj.diff_reports_by_location(report1_location, report2_location, diff_output_directory, None):
    print 'Error encountered during diff'
  print 'Please inspect the generated reports manually'



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
import logging
import os
import glob
import sys
from nose import *

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))), 'src')))

import naarad.graphing.matplotlib_naarad as mpl_n
from naarad.graphing.plot_data import PlotData as PlotData
logger = logging.getLogger('naarad')

def validate(filename):
  if os.path.exists(filename):
    if not os.path.getsize(filename):
      return False
    else:
      return True
  else:
    return False

def plot(list_of_plots, output_file, scenario_name, validation):
  mpl_n.graph_data(list_of_plots, '.', output_file)
  assert validate(output_file + '.png') == validation

def init_logging(log_level):
  log_file = 'test_matplotlib.log'
  # clear the log file
  with open(log_file, 'w'):
    pass
  numeric_level = getattr(logging, log_level.upper(), None) if log_level else logging.INFO
  if not isinstance(numeric_level, int):
    raise ValueError('Invalid log level: %s' % log_level)
  logger.setLevel(logging.DEBUG)
  fh = logging.FileHandler(log_file)
  fh.setLevel(logging.DEBUG)
  ch = logging.StreamHandler()
  ch.setLevel(numeric_level)
  formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
  fh.setFormatter(formatter)
  ch.setFormatter(formatter)
  logger.addHandler(fh)
  logger.addHandler(ch)

def setup_matplotlib_tests():
  init_logging('INFO')
  png_list = glob.glob('test*.png')
  for png in png_list:
    os.remove(png)
    logger.info('Deleting : %s', png)

#Test Cases
@with_setup(setup_matplotlib_tests())
def test_generator():
  #Test Data
  plot1 = PlotData('test1.csv', 1, 'GC Commit', 'MB', None, 600, 1200, 'line')
  plot2 = PlotData('test2.csv', 1, 'GC CMS Pause', 'seconds', None, 600, 1200, 'line')
  plot3 = PlotData('test3.csv', 1, 'GC Promo', 'bps', None, 600, 1200, 'line')
  plot4 = PlotData('test4.csv', 1, 'GC Promo', 'bps', None, 600, 1200, 'line')
  plot5 = PlotData('test5.csv', 1, 'GC Promo', 'bps', None, 600, 1200, 'line')
  yield plot, [plot1], 'test1', 'single plot with all valid csv', True
  yield plot, [plot1, plot2], 'test2', 'dual plot with all valid csv', True
  yield plot, [plot1, plot2, plot3], 'test3', 'multi plot with all valid csv', True
  yield plot, [plot4], 'test4', 'single plot with 1 empty csv', False
  yield plot, [plot4, plot1], 'test5', 'dual plot with 1 empty csv', True
  yield plot, [plot1, plot4, plot3], 'test6', 'multi plot with 1 empty csv', True
  yield plot, [plot5], 'test7', 'single plot with 1 non-existent csv', False
  yield plot, [plot5, plot1], 'test8', 'dual plot with 1 non-existent csv', True
  yield plot, [plot1, plot5, plot3], 'test9', 'multi plot with 1 non-existent csv', True


# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
from collections import defaultdict
import datetime
import gc
import logging
import os
import re
import numpy
from naarad.metrics.metric import Metric
import naarad.utils

logger = logging.getLogger('naarad.metrics.ProcMeminfoMetric')

class ProcMeminfoMetric(Metric):
  """
  logs of /proc/vmstat
  The raw log file is assumed to have a timestamp prefix of all lines. E.g. in the format of "2013-01-02 03:55:22.13456 compact_fail 36"
  The log lines can be generated by   'cat /proc/vmstat | sed "s/^/$(date +%Y-%m-%d\ %H:%M:%S.%05N)\t/" '  
  """

  sub_metrics = None
  
  def __init__ (self, metric_type, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end,
                rule_strings, important_sub_metrics, **other_options):
    Metric.__init__(self, metric_type, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end,
                    rule_strings, important_sub_metrics)

    # in particular, Section can specify a subset of all rows (default has 43 rows):  "sub_metrics=nr_free_pages nr_inactive_anon"
    for (key, val) in other_options.iteritems():
      setattr(self, key, val.split())   
      
    self.sub_metric_description = {
      'MemTotal': 'Total memory in KB',
      'MemFree': 'Total free memory in KB',
      'Buffers': 'Size of buffers in KB',
      'Cached': 'Size of page cache in KB',
     }    
     
 
  def parse(self):
    """
    Parse the vmstat file
    :return: status of the metric parse
    """
    file_status = True
    for input_file in self.infile_list:
      file_status = file_status and naarad.utils.is_valid_file(input_file)
      if not file_status:
        return False
    status = True
    data = {}  # stores the data of each column
    for input_file in self.infile_list:
      logger.info('Processing : %s',input_file)
      with open(input_file) as fh:
        for line in fh:
          words = line.split()        # [0] is day; [1] is seconds; [2] is field name:; [3] is value  [4] is unit
          if len(words) < 3:
            continue
          ts = words[0] + " " + words[1]
          if self.ts_out_of_range(ts):
            continue
          col = words[2].strip(':')
          # only process sub_metrics specified in config.
          if self.sub_metrics and col not in self.sub_metrics:
            continue
          # add unit to metric description; most of the metrics have 'KB'; a few others do not have unit, they are in number of pages
          if len(words) > 4 and words[4]:
            unit = words[4]
          else:
            unit = 'pages'
          self.sub_metric_unit[col] = unit
          # stores the values in data[] before finally writing out
          if col in self.column_csv_map:
            out_csv = self.column_csv_map[col]
          else:
            out_csv = self.get_csv(col)   #  column_csv_map[] is assigned in get_csv()
            data[out_csv] = []
          data[out_csv].append(ts + "," + words[3])
    #post processing, putting data in csv files;   
    for csv in data.keys():      
      self.csv_files.append(csv)
      with open(csv, 'w') as fh:
        fh.write('\n'.join(sorted(data[csv])))
    return status



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");?you may not use this file except in compliance with the
License.?You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software?distributed under the License is distributed on an
"AS IS" BASIS,?WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""

from collections import defaultdict
import ConfigParser
import errno
import logging
import os
import threading
import naarad.utils
import naarad.naarad_constants as CONSTANTS
from naarad_imports import metric_classes
from naarad_imports import aggregate_metric_classes
from naarad_imports import graphing_modules
from naarad_imports import reporting_modules
from naarad.reporting.diff import Diff
from naarad.reporting.diff import NaaradReport

logger = logging.getLogger('naarad')

class _Analysis(object):
  """
  Class that saves state for analysis to be conducted
  """
  def __init__(self, ts_start, config, test_id=None):
    self.ts_start = ts_start
    self.ts_end = None
    self.test_id = test_id
    self.config = config
    self.description = ''
    self.input_directory = None
    self.output_directory = None
    self.resource_path = 'resources'
    self.status = CONSTANTS.OK
    self.sla_data = {}
    self.stats_data = {}
    self.variables = None

class Naarad(object):
  """
  Naarad base class that will let the caller run multiple naarad analysis
  """

  def __init__(self):
    self._default_test_id = -1
    self._analyses = {}
    self._resource_path = 'resources'
    self._input_directory = None
    self._output_directory = None
    self.return_exit_code = False
    self.skip_plots = False
    self.available_graphing_modules = graphing_modules
    logger.info('Available graphing modules: %s ', ','.join(self.available_graphing_modules.keys()))
    naarad.metrics.metric.Metric.graphing_modules = self.available_graphing_modules
    naarad.reporting.diff.Diff.graphing_modules = self.available_graphing_modules
    naarad.metrics.metric.Metric.device_types = CONSTANTS.device_type_metrics

  def signal_start(self, config, test_id=None, **kwargs):
    """
    Initialize an analysis object and set ts_start for the analysis represented by test_id
    :param test_id: integer that represents the analysis
    :param config: config can be a ConfigParser.ConfigParser object or a string specifying local or http(s) location
     for config
    :return: test_id
    """
    if not test_id:
      self._default_test_id += 1
      test_id = self._default_test_id
    self._analyses[test_id] = _Analysis(naarad.utils.get_standardized_timestamp('now', None), config,
                                      test_id=test_id)
    if kwargs:
      if 'description' in kwargs.keys():
        self._analyses[test_id].description = kwargs['description']
      if 'input_directory' in kwargs.keys():
        self._analyses[test_id].input_directory = kwargs['input_directory']
      if 'output_directory' in kwargs.keys():
        self._analyses[test_id].output_directory = kwargs['output_directory']
    return test_id

  def signal_stop(self, test_id=None):
    """
    Set ts_end for the analysis represented by test_id
    :param test_id: integer that represents the analysis
    :return: test_id
    """
    if test_id is None:
      test_id = self._default_test_id
    if self._analyses[test_id].ts_end:
      return CONSTANTS.OK
    self._analyses[test_id].ts_end = naarad.utils.get_standardized_timestamp('now', None)
    return CONSTANTS.OK

  def get_failed_analyses(self):
    """
    Returns a list of test_id for which naarad analysis failed
    :return: list of test_ids
    """
    failed_analyses = []
    for test_id in self._analyses.keys():
      if self._analyses[test_id].status != CONSTANTS.OK:
        failed_analyses.append(test_id)
    return failed_analyses

  def get_sla_data(self, test_id):
    """
    Returns sla data for all the metrics associated with a test_id
    :return: dict of form { metric.label:metric.sla_map}
    """
    return self._analyses[test_id].sla_data

  def _set_sla_data(self, test_id, metrics):
    """
    Get sla data from each metric and set it in the _Analysis object specified by test_id to make it available
    for retrieval
    :return: currently always returns CONSTANTS.OK. Maybe enhanced in future to return additional status
    """
    for metric in metrics:
      self._analyses[test_id].sla_data[metric.label] = metric.sla_map
    return CONSTANTS.OK

  def get_stats_data(self, test_id):
    """
    Returns summary stats data for all the metrics associated with a test_id
    :return: dict of form { metric.label:metric.summary_stats}
    """
    return self._analyses[test_id].stats_data

  def _set_stats_data(self, test_id, metrics):
    """
    Get summary stats data from each metric and set it in the _Analysis object specified by test_id to make it available
    for retrieval
    :return: currently always returns CONSTANTS.OK. Maybe enhanced in future to return additional status
    """
    for metric in metrics:
      self._analyses[test_id].stats_data[metric.label] = metric.summary_stats
    return CONSTANTS.OK

  def _create_output_directories(self, analysis):
    """
    Create the necessary output and resource directories for the specified analysis
    :param: analysis: analysis associated with a given test_id
    """
    try:
      os.makedirs(analysis.output_directory)
    except OSError as exception:
      if exception.errno != errno.EEXIST:
        raise
    try:
      resource_directory = os.path.join(analysis.output_directory, analysis.resource_path)
      os.makedirs(resource_directory)
    except OSError as exception:
      if exception.errno != errno.EEXIST:
        raise

  def _run_pre(self, analysis, run_steps):
    """
    If Naarad is run in CLI mode, execute any pre run steps specified in the config. ts_start/ts_end are set based on
    workload run steps if any.
    :param: analysis: The analysis object being processed
    :param: run_steps: list of post run steps
    """
    workload_run_steps = []
    for run_step in sorted(run_steps, key=lambda step: step.run_rank):
      run_step.run()
      if run_step.run_type == CONSTANTS.RUN_TYPE_WORKLOAD:
        workload_run_steps.append(run_step)
    # Get analysis time period from workload run steps
    if len(workload_run_steps) > 0:
      analysis.ts_start, analysis.ts_end = naarad.utils.get_run_time_period(workload_run_steps)
    return CONSTANTS.OK

  def _run_post(self, run_steps):
    """
    If Naarad is run in CLI mode, execute any post run steps specified in the config
    :param: run_steps: list of post run steps
    """
    for run_step in sorted(run_steps, key=lambda step: step.run_rank):
      run_step.run()
    return CONSTANTS.OK

  def _process_args(self, analysis, args):
    """
    When Naarad is run in CLI mode, get the CL arguments and update the analysis
    :param: analysis: The analysis being processed
    :param: args: Command Line Arguments received by naarad
    """
    if args.exit_code:
      self.return_exit_code = args.exit_code
    if args.no_plots:
      self.skip_plots = args.no_plots
    if args.start:
      analysis.ts_start = naarad.utils.get_standardized_timestamp(args.start, None)
    if args.end:
      analysis.ts_end = naarad.utils.get_standardized_timestamp(args.end, None)
    if args.variables:
      analysis.variables = naarad.utils.get_variables(args)
    return CONSTANTS.OK

  def analyze(self, input_directory, output_directory, **kwargs):
    """
    Run all the analysis saved in self._analyses, sorted by test_id
    :param: input_directory: location of log files
    :param: output_directory: root directory for analysis output
    :param: **kwargs: Optional keyword args
    :return: int: status code.
    """
    is_api_call = True
    if len(self._analyses) == 0:
      if 'config' not in kwargs.keys():
        return CONSTANTS.ERROR
      self._analyses[0] = _Analysis(None, kwargs['config'], test_id=0)
    if 'args' in kwargs:
      self._process_args(self._analyses[0], kwargs['args'])
      is_api_call = False
    error_count = 0
    self._input_directory = input_directory
    self._output_directory = output_directory
    for test_id in sorted(self._analyses.keys()):
      if not self._analyses[test_id].input_directory:
        self._analyses[test_id].input_directory = input_directory
      if not self._analyses[test_id].output_directory:
        if len(self._analyses) > 1:
          self._analyses[test_id].output_directory = os.path.join(output_directory, str(test_id))
        else:
          self._analyses[test_id].output_directory = output_directory
      if('config' in kwargs.keys()) and (not self._analyses[test_id].config):
        self._analyses[test_id].config = kwargs['config']
      self._create_output_directories(self._analyses[test_id])
      self._analyses[test_id].status = self.run(self._analyses[test_id], is_api_call, **kwargs)
      if self._analyses[test_id].status != CONSTANTS.OK:
        error_count += 1
    if error_count > 0:
      return CONSTANTS.ERROR
    else:
      return CONSTANTS.OK

  def run(self, analysis, is_api_call, **kwargs):
    """
    :param analysis: Run naarad analysis for the specified analysis object
    :param **kwargs: Additional keyword args can be passed in here for future enhancements
    :return:
    """
    threads = []
    crossplots = []
    report_args = {}
    metrics = defaultdict()
    run_steps = defaultdict(list)
    discovery_mode = False
    graphing_library = CONSTANTS.DEFAULT_GRAPHING_LIBRARY
    graph_timezone = None

    if isinstance(analysis.config, str):
      if not naarad.utils.is_valid_file(analysis.config):
        return CONSTANTS.INVALID_CONFIG
      config_object = ConfigParser.ConfigParser(analysis.variables)
      config_object.optionxform = str
      config_object.read(analysis.config)
    elif isinstance(analysis.config, ConfigParser.ConfigParser):
      config_object = analysis.config
    else:
      if is_api_call:
        return CONSTANTS.INVALID_CONFIG
      else:
        metrics['metrics'] = naarad.utils.discover_by_name(analysis.input_directory, analysis.output_directory)
        if len(metrics['metrics']) == 0:
          logger.warning('Unable to auto detect metrics in the specified input directory: %s', analysis.input_directory)
          return CONSTANTS.ERROR
        else:
          discovery_mode = True
          metrics['aggregate_metrics'] = []
    if not discovery_mode:
      metrics, run_steps, crossplots, report_args, graph_timezone, graphing_library = self._process_naarad_config(config_object, analysis)

    # If graphing libraries are not installed, skip static images
    if not graphing_library in self.available_graphing_modules.keys():
      logger.error("Naarad cannot import graphing library %s on your system. Will not generate static charts", graphing_library)
      self.skip_plots = True

    if not is_api_call:
      self._run_pre(analysis, run_steps['pre'])
    for metric in metrics['metrics']:
      if analysis.ts_start:
        metric.ts_start = analysis.ts_start
      if analysis.ts_end:
        metric.ts_end = analysis.ts_end
      thread = threading.Thread(target=naarad.utils.parse_and_plot_single_metrics, args=(metric, graph_timezone, analysis.output_directory, analysis.input_directory, graphing_library, self.skip_plots))
      thread.start()
      threads.append(thread)
    for t in threads:
      t.join()
    for metric in metrics['aggregate_metrics']:
      thread = threading.Thread(target=naarad.utils.parse_and_plot_single_metrics, args=(metric, graph_timezone, analysis.output_directory, analysis.input_directory, graphing_library, self.skip_plots))
      thread.start()
      threads.append(thread)
    for t in threads:
      t.join()
    self._set_sla_data(analysis.test_id, metrics['metrics'] + metrics['aggregate_metrics'])
    self._set_stats_data(analysis.test_id, metrics['metrics'] + metrics['aggregate_metrics'])
    if len(crossplots) > 0 and not self.skip_plots:
      correlated_plots = naarad.utils.nway_plotting(crossplots, metrics['metrics'] + metrics['aggregate_metrics'],
                                                    os.path.join(analysis.output_directory, analysis.resource_path),
                                                    analysis.resource_path, graphing_library)
    else:
      correlated_plots = []
    rpt = reporting_modules['report'](None, analysis.output_directory, os.path.join(analysis.output_directory, analysis.resource_path), analysis.resource_path, metrics['metrics'] + metrics['aggregate_metrics'], correlated_plots=correlated_plots, **report_args)
    rpt.generate()
    if not is_api_call:
      self._run_post(run_steps['post'])

    if self.return_exit_code:
      for metric in metrics['metrics'] + metrics['aggregate_metrics']:
        if metric.status == CONSTANTS.SLA_FAILED:
          return CONSTANTS.SLA_FAILURE

    return CONSTANTS.OK

  def diff(self, test_id_1, test_id_2, config=None, **kwargs):
    """
    Create a diff report using test_id_1 as a baseline
    :param: test_id_1: test id to be used as baseline
    :param: test_id_2: test id to compare against baseline
    :param: config file for diff (optional)
    :param: **kwargs: keyword arguments
    """
    output_directory = os.path.join(self._output_directory,'diff_' + str(test_id_1) + '_' + str(test_id_2))
    if kwargs:
      if 'output_directory' in kwargs.keys():
        output_directory = kwargs['output_directory']
    diff_report = Diff([NaaradReport(self._analyses[test_id_1].output_directory, None),
                        NaaradReport(self._analyses[test_id_2].output_directory, None)],
                       'diff', output_directory, os.path.join(output_directory, self._resource_path),
                       self._resource_path)
    if config:
      naarad.utils.extract_diff_sla_from_config_file(diff_report, config)
    diff_report.generate()
    if diff_report.sla_failures > 0:
      return CONSTANTS.SLA_FAILURE
    if diff_report.status != 'OK':
      return CONSTANTS.ERROR
    return CONSTANTS.OK

  def diff_reports_by_location(self, report1_location, report2_location, output_directory, config=None, **kwargs):
    """
    Create a diff report using report1 as a baseline
    :param: report1_location: report to be used as baseline
    :param: report2_location: report to compare against baseline
    :param: config file for diff (optional)
    :param: **kwargs: keyword arguments
    """

    if kwargs:
      if 'output_directory' in kwargs.keys():
        output_directory = kwargs['output_directory']
    diff_report = Diff([NaaradReport(report1_location, None), NaaradReport(report2_location, None)], 'diff',
                       output_directory, os.path.join(output_directory, self._resource_path), self._resource_path)
    if config:
      naarad.utils.extract_diff_sla_from_config_file(diff_report, config)
    diff_report.generate()
    if diff_report.sla_failures > 0:
      return CONSTANTS.SLA_FAILURE
    if diff_report.status != 'OK':
      return CONSTANTS.ERROR
    return CONSTANTS.OK


  def _process_naarad_config(self, config, analysis):
    """
    Process the config file associated with a particular analysis and return metrics, run_steps and crossplots.
    Also sets output directory and resource_path for an anlaysis
    """
    graph_timezone = None
    output_directory = analysis.output_directory
    resource_path = analysis.resource_path
    run_steps = defaultdict(list)
    metrics = defaultdict(list)
    indir_default = ''
    crossplots = []
    report_args = {}
    graphing_library = None

    if config.has_section('GLOBAL'):
      ts_start, ts_end = naarad.utils.parse_global_section(config, 'GLOBAL')
      if config.has_option('GLOBAL', 'user_defined_metrics'):
        naarad.utils.parse_user_defined_metric_classes(config, metric_classes)
      config.remove_section('GLOBAL')

    if config.has_section('REPORT'):
      report_args = naarad.utils.parse_report_section(config, 'REPORT')
      config.remove_section('REPORT')

    for section in config.sections():
      # GRAPH section is optional
      if section == 'GRAPH':
        graphing_library, crossplots, outdir_default, indir_default, graph_timezone = \
          naarad.utils.parse_graph_section(config, section, output_directory, indir_default)
      elif section.startswith('RUN-STEP'):
        run_step = naarad.utils.parse_run_step_section(config, section)
        if not run_step:
          logger.error('Ignoring section %s, could not parse it correctly', section)
          continue
        if run_step.run_order == CONSTANTS.PRE_ANALYSIS_RUN:
          run_steps['pre'].append(run_step)
        # DURING_ANALYSIS_RUN not supported yet
        elif run_step.run_order == CONSTANTS.DURING_ANALYSIS_RUN:
          run_steps['in'].append(run_step)
        elif run_step.run_order == CONSTANTS.POST_ANALYSIS_RUN:
          run_steps['post'].append(run_step)
        else:
          logger.error('Unknown RUN-STEP run_order specified')
      else:
        # section name is used to create sub-directories, so enforce it.
        if not naarad.utils.is_valid_metric_name(section):
          logger.critical('Section name %s is invalid! Only letters, digits, dot(.), dash(-), underscore(_) are allowed'
                          % section)
          return CONSTANTS.CRITICAL_FAILURE
        if section == 'SAR-*':
          hostname, infile, label, ts_start, ts_end, precision, kwargs, rule_strings = \
            naarad.utils.parse_basic_metric_options(config, section)
          sar_metrics = naarad.utils.get_all_sar_objects(metrics, infile, hostname, output_directory, label, ts_start,
                                                         ts_end, None)
          metrics['metrics'].extend(sar_metrics)
        else:
          new_metric = naarad.utils.parse_metric_section(config, section, metric_classes, metrics['metrics'],
                                                         aggregate_metric_classes, output_directory, resource_path)
          new_metric.bin_path = os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(
            os.path.dirname(os.path.abspath(__file__)))),'bin'))
          metric_type = section.split('-')[0]
          if metric_type in aggregate_metric_classes:
            metrics['aggregate_metrics'].append(new_metric)
          else:
            metrics['metrics'].append(new_metric)
    return metrics, run_steps, crossplots, report_args, graph_timezone, graphing_library



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
from collections import defaultdict
import datetime
import gc
import logging
import os
import re
import numpy
from naarad.metrics.metric import Metric
import naarad.utils

logger = logging.getLogger('naarad.metrics.ProcVmstatMetric')

class ProcVmstatMetric(Metric):
  """
  logs of /proc/vmstat
  The raw log file is assumed to have a timestamp prefix of all lines. E.g. in the format of "2013-01-02 03:55:22.13456 compact_fail 36"
  The log lines can be generated by   'cat /proc/vmstat | sed "s/^/$(date +%Y-%m-%d\ %H:%M:%S.%05N)\t/" '  
  """
  
  def __init__ (self, metric_type, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end,
                rule_strings, important_sub_metrics, **other_options):
    Metric.__init__(self, metric_type, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end,
                    rule_strings, important_sub_metrics)

    self.sub_metrics = None
    # in particular, Section can specify a subset of all rows (default has 86 rows):  "sub_metrics=nr_free_pages nr_inactive_anon"
    
    for (key, val) in other_options.iteritems():
      setattr(self, key, val.split())   
      
    self.sub_metric_description = {
      'nr_free_pages': 'Number of free pages',
      'nr_inactive_anon': 'Number of inactive anonymous pages',
      'nr_active_anon': 'Number of active anonymous pages',
      'nr_inactive_file': 'Number of inactive file pages',
      'nr_active_file': 'Number of active file pages',
     }

      
  def parse(self):
    """
    Parse the vmstat file
    :return: status of the metric parse
    """
    file_status = True
    for input_file in self.infile_list:
      file_status = file_status and naarad.utils.is_valid_file(input_file)
    if not file_status:
      return False
    status = True
    data = {}  # stores the data of each column
    for input_file in self.infile_list:
      logger.info('Processing : %s', input_file)
      with open(input_file) as fh:
        for line in fh:
          words = line.split()          # [0] is day; [1] is seconds; [2] is field name; [3] is value
          if len(words) < 3:
            continue
          ts = words[0] + " " + words[1]
          if self.ts_out_of_range(ts):
            continue
          col = words[2]
          # if sub_metrics is specified, only process those specified in config.
          if self.sub_metrics and col not in self.sub_metrics:
            continue
          self.sub_metric_unit[col] = 'pages'  # The unit of the metric. For /proc/vmstat, they are all in pages
          if col in self.column_csv_map:
            out_csv = self.column_csv_map[col]
          else:
            out_csv = self.get_csv(col)   #  column_csv_map[] is assigned in get_csv()
            data[out_csv] = []
          data[out_csv].append(ts + "," + words[3])
    #post processing, putting data in csv files;   
    for csv in data.keys():      
      self.csv_files.append(csv)
      with open(csv, 'w') as fh:
        fh.write('\n'.join(data[csv]))
    return status



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
import argparse
import calendar
import ConfigParser
import datetime
import imp
import logging
import numpy
import os
import pytz
from pytz import timezone
import re
import sys
import time
import urllib
from naarad.naarad_imports import metric_classes, aggregate_metric_classes
from naarad.sla import SLA
from naarad.metrics.sar_metric import SARMetric
from naarad.metrics.metric import Metric
from naarad.graphing.plot_data import PlotData
from naarad.run_steps.local_cmd import Local_Cmd
import naarad.naarad_constants as CONSTANTS

logger = logging.getLogger('naarad.utils')

def import_modules(module_dict, is_class_type=True):
  return_dict = {}
  for module_name, module_string in module_dict.items():
    try:
      if is_class_type:
        file_name, class_name = module_string.rsplit('.', 1)
        mod = __import__(file_name, fromlist=[class_name])
        return_dict[module_name] = getattr(mod, class_name)
      else:
        return_dict[module_name] = __import__(module_string, fromlist=[module_string])
    except ImportError:
      pass
  return return_dict

def parse_user_defined_metric_classes(config_obj, metric_classes):
  """
  Parse the user defined metric class information
  :param config_obj: ConfigParser object
  :param metric_classes: list of metric classes to be updated
  :return:
  """
  user_defined_metric_list = config_obj.get('GLOBAL', 'user_defined_metrics').split()
  for udm_string in user_defined_metric_list:
    try:
      metric_name, metric_class_name, metric_file = udm_string.split(':')
    except ValueError:
      logger.error('Bad user defined metric specified')
      continue
    module_name = os.path.splitext(os.path.basename(metric_file))[0]
    try:
      new_module = imp.load_source(module_name, metric_file)
      new_class = getattr(new_module, metric_class_name)
      if metric_name in metric_classes.keys():
        logger.warn('Overriding pre-defined metric class definition for ', metric_name)
      metric_classes[metric_name] = new_class
    except ImportError:
      logger.error('Something wrong with importing a user defined metric class. Skipping metric: ', metric_name)
      continue

def is_valid_url(url):
  """
  Check if a given string is in the correct URL format or not

  :param str url:
  :return: True or False
  """
  regex = re.compile(
      r'^(http|https|ftp)://[A-Za-z0-9]+(-[A-Za-z0-9]+)*([:][A-Za-z0-9]+(-[A-Za-z0-9]+)*){0,1}([@][A-Za-z0-9]+(-[A-Za-z0-9]+)*){0,1}(\.[A-Za-z0-9]+(-[A-Za-z0-9]+)*)*(:[0-9]{1,5}){0,1}(/[A-Za-z0-9]*[A-Za-z0-9-._\(\)]*)*([?#&].*)*$')
  if regex.match(url):
    logger.info( "URL given as config")
    return True
  else:
    return False

def download_file(url):
  """
  Download a file pointed to by url to a temp file on local disk

  :param str url:
  :return: local_file
  """
  try:
    (local_file, headers) = urllib.urlretrieve(url)
  except:
    sys.exit("ERROR: Problem downloading config file. Please check the URL (" + url + "). Exiting...")
  return local_file

def sanitize_string_section_name(string):
  string = string.replace('/', '_')
  string = string.replace('%', '_')
  return string

def is_valid_metric_name(metric_name):
  """
  check the validity of metric_name in config; the metric_name will be used for creation of sub-dir, so only contains: alphabet, digits , '.', '-' and '_'
  :param str metric_name: metric_name
  :return: True if valid
  """
  reg=re.compile('^[a-zA-Z0-9\.\-\_]+$')
  if reg.match(metric_name) and not metric_name.startswith('.'):
    return True
  else:
    return False

def get_run_time_period(run_steps):
  """
  This method finds the time range which covers all the Run_Steps

  :param run_steps: list of Run_Step objects
  :return: tuple of start and end timestamps
  """
  init_ts_start = get_standardized_timestamp('now', None)
  ts_start = init_ts_start
  ts_end = '0'
  for run_step in run_steps:
    if run_step.ts_start and run_step.ts_end:
      if run_step.ts_start < ts_start:
        ts_start = run_step.ts_start
      if run_step.ts_end > ts_end:
        ts_end = run_step.ts_end
  if ts_end == '0':
    ts_end = None
  if ts_start == init_ts_start:
    ts_start = None
  logger.info('get_run_time_period range returned ' + str(ts_start) + ' to ' + str(ts_end))
  return ts_start, ts_end

def get_rule_strings(config_obj, section):
  """
  Extract rule strings from a section
  :param config_obj: ConfigParser object
  :param section: Section name
  :return: the rule strings
  """
  rule_strings = {}
  kwargs = dict(config_obj.items(section))
  for key in kwargs.keys():
    if key.endswith('.sla'):
      rule_strings[key.replace('.sla','')] = kwargs[key]
      del kwargs[key]
  return rule_strings, kwargs

def extract_diff_sla_from_config_file(obj, options_file):
  """
  Helper function to parse diff config file, which contains SLA rules for diff comparisons
  """
  rule_strings = {}
  config_obj = ConfigParser.ConfigParser()
  config_obj.optionxform = str
  config_obj.read(options_file) 
  for section in config_obj.sections():
    rule_strings, kwargs = get_rule_strings(config_obj, section)
    for (key, val) in rule_strings.iteritems():
      set_sla(obj, section, key, val)

def parse_basic_metric_options(config_obj, section):
  """
  Parse basic options from metric sections of the config
  :param config_obj: ConfigParser object
  :param section: Section name
  :return: all the parsed options
  """
  infile = {}
  aggr_hosts = None
  aggr_metrics = None
  ts_start = None
  ts_end = None
  precision = None
  hostname = "localhost"
  rule_strings = {}
  important_sub_metrics = None

  try:
    if config_obj.has_option(section, 'important_sub_metrics'):
      important_sub_metrics = config_obj.get(section, 'important_sub_metrics').split()
      config_obj.remove_option(section, 'important_sub_metrics')

    if config_obj.has_option(section, 'hostname'):
      hostname = config_obj.get(section, 'hostname')
      config_obj.remove_option(section, 'hostname')
    
    #'infile' is not mandatory for aggregate metrics
    if config_obj.has_option(section,'infile'):
      infile = config_obj.get(section, 'infile').split()
      config_obj.remove_option(section, 'infile')

    label = sanitize_string_section_name(section)
    if config_obj.has_option(section, 'ts_start'):
      ts_start = get_standardized_timestamp(config_obj.get(section, 'ts_start'), None)
      config_obj.remove_option(section, 'ts_start')
    if config_obj.has_option(section, 'ts_end'):
      ts_end = get_standardized_timestamp(config_obj.get(section, 'ts_end'), None)
      config_obj.remove_option(section, 'ts_end')
    if config_obj.has_option(section, 'precision'):
      precision = config_obj.get(section, 'precision')
      config_obj.remove_option(section, 'precision')
    #support aggregate metrics, which take aggr_hosts and aggr_metrics
    if config_obj.has_option(section, 'aggr_hosts'):
      aggr_hosts = config_obj.get(section, 'aggr_hosts')
      config_obj.remove_option(section, 'aggr_hosts')
    if config_obj.has_option(section, 'aggr_metrics'):
      aggr_metrics = config_obj.get(section, 'aggr_metrics')
      config_obj.remove_option(section, 'aggr_metrics')
    rule_strings, other_options = get_rule_strings(config_obj, section)
  except ConfigParser.NoOptionError:
    logger.exception("Exiting.... some mandatory options are missing from the config file in section: " + section)
    sys.exit()
  return hostname, infile, aggr_hosts, aggr_metrics, label, ts_start, ts_end, precision, other_options, rule_strings, \
         important_sub_metrics

def parse_metric_section(config_obj, section, metric_classes,  metrics, aggregate_metric_classes, outdir_default, resource_path):
  """
  Parse a metric section and create a Metric object
  :param config_obj: ConfigParser object
  :param section: Section name
  :param metric_classes: List of valid metric types
  :param metrics: List of all regular metric objects (used by aggregate metric)
  :param aggregate_metric_classes: List of all valid aggregate metric types
  :param outdir_default: Default output directory
  :param resource_path: Default resource directory
  :return: An initialized Metric object
  """
  hostname, infile, aggr_hosts, aggr_metrics, label, ts_start, ts_end, precision, other_options, rule_strings, \
    important_sub_metrics = parse_basic_metric_options(config_obj, section)

  #TODO: Make user specify metric_type in config and not infer from section
  metric_type = section.split('-')[0]
  if metric_type in aggregate_metric_classes:
    new_metric = initialize_aggregate_metric(section, aggr_hosts, aggr_metrics, metrics, outdir_default, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics, other_options)
  else:
    new_metric = initialize_metric(section, infile , hostname, outdir_default, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics, other_options)

  if config_obj.has_option(section, 'ignore') and config_obj.getint(section, 'ignore') == 1:
    new_metric.ignore = True
  if config_obj.has_option(section, 'calc_metrics'):
    new_metric.calc_metrics = config_obj.get(section, 'calc_metrics')
  new_metric.precision = precision
  return new_metric

def parse_global_section(config_obj, section):
  """
  Parse GLOBAL section in the config to return global settings
  :param config_obj: ConfigParser object
  :param section: Section name
  :return: ts_start and ts_end time
  """
  ts_start = None
  ts_end = None
  if config_obj.has_option(section, 'ts_start'):
    ts_start = get_standardized_timestamp(config_obj.get(section, 'ts_start'), None)
    config_obj.remove_option(section, 'ts_start')
  if config_obj.has_option(section, 'ts_end'):
    ts_end = get_standardized_timestamp(config_obj.get(section, 'ts_end'), None)
    config_obj.remove_option(section, 'ts_end')
  return ts_start, ts_end

def parse_run_step_section(config_obj, section):
  """
  Parse a RUN-STEP section in the config to return a Run_Step object
  :param config_obj: ConfigParser objection
  :param section: Section name
  :return: an initialized Run_Step object
  """
  kill_after_seconds = None
  try:
    run_cmd = config_obj.get(section, 'run_cmd')
    run_rank = int(config_obj.get(section, 'run_rank'))
  except ConfigParser.NoOptionError:
    logger.exception("Exiting.... some mandatory options are missing from the config file in section: " + section)
    sys.exit()
  except ValueError:
    logger.error("Bad run_rank %s specified in section %s, should be integer. Exiting.", config_obj.get(section, 'run_rank'), section)
    sys.exit()
  if config_obj.has_option(section, 'run_type'):
    run_type = config_obj.get(section, 'run_type')
  else:
    run_type = CONSTANTS.RUN_TYPE_WORKLOAD
  if config_obj.has_option(section, 'run_order'):
    run_order = config_obj.get(section, 'run_order')
  else:
    run_order = CONSTANTS.PRE_ANALYSIS_RUN
  if config_obj.has_option(section, 'call_type'):
    call_type = config_obj.get(section, 'call_type')
  else:
    call_type = 'local'
  if config_obj.has_option(section, 'kill_after_seconds'):
    try:
      kill_after_seconds = int(config_obj.get(section, 'kill_after_seconds'))
    except ValueError:
      logger.error("Bad kill_after_seconds %s specified in section %s, should be integer.", config_obj.get(section, 'kill_after_seconds'), section)

  if call_type == 'local':
    run_step_obj = Local_Cmd(run_type, run_cmd, call_type, run_order, run_rank, kill_after_seconds=kill_after_seconds)
  else:
    logger.error('Unsupported RUN_STEP supplied, call_type should be local')
    run_step_obj = None
  return run_step_obj

def parse_graph_section(config_obj, section, outdir_default, indir_default):
  """
  Parse the GRAPH section of the config to extract useful values
  :param config_obj: ConfigParser object
  :param section: Section name
  :param outdir_default: Default output directory passed in args
  :param indir_default: Default input directory passed in args
  :return: List of options extracted from the GRAPH section
  """
  graph_timezone = None
  graphing_library = CONSTANTS.DEFAULT_GRAPHING_LIBRARY
  crossplots = []

  if config_obj.has_option(section, 'graphing_library'):
    graphing_library = config_obj.get(section, 'graphing_library')
  if config_obj.has_option(section, 'graphs'):
    graphs_string = config_obj.get(section, 'graphs')
    crossplots = graphs_string.split()
    # Supporting both outdir and output_dir
  if config_obj.has_option(section, 'outdir'):
    outdir_default = config_obj.get(section, 'outdir')
  if config_obj.has_option(section, 'output_dir'):
    outdir_default = config_obj.get(section, 'output_dir')
  if config_obj.has_option(section, 'input_dir'):
    indir_default = config_obj.get(section, 'input_dir')
  if config_obj.has_option(section, 'graph_timezone'):
    graph_timezone = config_obj.get(section, 'graph_timezone')
    if graph_timezone not in ("UTC", "PST", "PDT"):
      logger.warn('Unsupported timezone ' + graph_timezone + ' specified in option graph_timezone. Will use UTC instead')
      graph_timezone = "UTC"
  return graphing_library, crossplots, outdir_default, indir_default, graph_timezone

def parse_report_section(config_obj, section):
  """
  parse the [REPORT] section of a config file to extract various reporting options to be passed to the Report object
  :param: config_obj : configparser object for the config file passed in to naarad
  :param: section: name of the section. 'REPORT' should be passed in here
  :return: report_kwargs: dictionary of Reporting options and values specified in config.
  """
  report_kwargs = {}
  if config_obj.has_option(section, 'stylesheet_includes'):
    report_kwargs['stylesheet_includes'] = config_obj.get(section, 'stylesheet_includes')
  if config_obj.has_option(section, 'javascript_includes'):
    report_kwargs['javascript_includes'] = config_obj.get(section, 'javascript_includes')
  if config_obj.has_option(section, 'header_template'):
    report_kwargs['header_template'] = config_obj.get(section, 'header_template')
  if config_obj.has_option(section, 'footer_template'):
    report_kwargs['footer_template'] = config_obj.get(section, 'footer_template')
  if config_obj.has_option(section, 'summary_content_template'):
    report_kwargs['summary_content_template'] = config_obj.get(section, 'summary_content_template')
  if config_obj.has_option(section, 'summary_page_template'):
    report_kwargs['summary_page_template'] = config_obj.get(section, 'summary_page_template')
  if config_obj.has_option(section, 'metric_page_template'):
    report_kwargs['metric_page_template'] = config_obj.get(section, 'metric_page_template')
  if config_obj.has_option(section, 'client_charting_template'):
    report_kwargs['client_charting_template'] = config_obj.get(section, 'client_charting_template')
  if config_obj.has_option(section, 'diff_client_charting_template'):
    report_kwargs['diff_client_charting_template'] = config_obj.get(section, 'diff_client_charting_template')
  if config_obj.has_option(section, 'diff_page_template'):
    report_kwargs['diff_page_template'] = config_obj.get(section, 'diff_page_template')
  return report_kwargs

def reconcile_timezones(begin_ts, ts_timezone, graph_timezone):
  if not graph_timezone:
    return begin_ts
  # Converting timestamp strings to be in timezone: graph_timezone
  # We only support UTC and PDT
  if graph_timezone != ts_timezone:
    utc = pytz.utc
    pst = timezone('US/Pacific')
    if graph_timezone == "UTC":
      # Assume timezone is PDT
      try:
        dt = pst.localize(datetime.datetime.strptime(begin_ts,"%Y-%m-%d %H:%M:%S"))
      except ValueError:
        dt = pst.localize(datetime.datetime.strptime(begin_ts,"%Y-%m-%d %H:%M:%S.%f"))
      begin_ts = dt.astimezone(utc).strftime("%Y-%m-%d %H:%M:%S.%f")
    else:
      # Assume timezone is UTC since graph_timezone is PDT
      try:
        dt = utc.localize(datetime.datetime.strptime(begin_ts,"%Y-%m-%d %H:%M:%S"))
      except ValueError:
        dt = utc.localize(datetime.datetime.strptime(begin_ts,"%Y-%m-%d %H:%M:%S.%f"))
      begin_ts = dt.astimezone(pst).strftime("%Y-%m-%d %H:%M:%S.%f")
  return begin_ts

def convert_to_unixts(ts_string):
  try:
    dt_obj = datetime.datetime.strptime(ts_string, "%Y-%m-%d %H:%M:%S.%f")
  except ValueError:
    dt_obj = datetime.datetime.strptime(ts_string, "%Y-%m-%d %H:%M:%S")
  return float(calendar.timegm(dt_obj.utctimetuple())*1000.0 + dt_obj.microsecond/1000.0)

def is_number(string):
  try:
    float(string)
    return True
  except ValueError:
    return False

def get_all_sar_objects(metrics, indir, hostname, output_directory, label, ts_start, ts_end, options):
  metrics = []
  sar_types = ('device', 'cpuusage', 'cpuhz', 'memory', 'memutil', 'paging')
  for sar_metric_type in sar_types:
    #infile = indir + '/' + 'sar.' + sar_metric_type + '.out'
    infile = os.path.join(indir, 'sar.' + sar_metric_type + '.out')
    if os.path.exists(infile):
      obj_type = 'SAR-' + sar_metric_type
      metric = SARMetric(obj_type, infile, hostname, output_directory, label, ts_start, ts_end, options)
      metrics.append(metric)
  return metrics

def sanitize_string(string):
  string = string.replace('/', '-per-')
  if string.startswith('%'):
    string = string.replace('%', 'percent-')
  else:
    string = string.replace('.%', '.percent-') #handle the cases of "all.%sys"
    string = string.replace('%', '-percent-')
  return string

def get_default_csv(output_directory, val):
  val = sanitize_string(val)
  return os.path.join(output_directory, val + '.csv')

def convert_to_24hr_format(ts):
  words = ts.split()
  if len(words) == 1:
    return ts
  if words[1] == 'PM':
    tmp = words[0].split(':')
    if tmp[0] != '12':
      hour = int(tmp[0]) + 12
      tmp[0] = str(hour)
    ts = ":".join(tmp)
  elif words[1] == 'AM':
    tmp = words[0].split(':')
    if tmp[0] == '12':
      tmp[0] = '00'
    ts = ":".join(tmp)
  return ts

def get_merged_csvname(output_directory, vals):
  return os.path.join(output_directory, '-'.join(vals) + '.csv')

def get_merged_charttitle(vals):
  return " vs ".join(vals)

def get_merged_plot_link_name(vals):
  return '-'.join(vals)

def get_merged_png_name(vals):
  # The merged_png_name format is relied upon by naarad.reporting.report.is_correlated_image method.
  return '-'.join(vals)

def generate_html_report(output_directory, html_string):
  htmlfilename = os.path.join(output_directory, 'Report.html')
  with open(htmlfilename, 'w') as htmlf:
    header = '<html><head><title>naarad analysis report</title>'
    dygraphs_include = '''<script type='text/javascript'
      src='http://dygraphs.com/dygraph-combined.js'></script>
      '''
    sorttable_include = '<script type="text/javascript" src="http://www.kryogenix.org/code/browser/sorttable/sorttable.js"></script>'
    body = '</head><body>'
    footer = '</body></html>'
    htmlf.write(header)
    htmlf.write(sorttable_include)
    htmlf.write(dygraphs_include)
    htmlf.write(body)
    htmlf.write(html_string)
    htmlf.write(footer)

def tscsv_nway_file_merge(outfile, filelist, filler):
  logger.info('called nway merge with %s', filelist)
  with open(outfile, 'w') as outf:
    filehandlers = [None] * len(filelist)
    currlines = [None] * len(filelist)
    for i in range(len(filelist)):
      try:
        filehandlers[i] = open(filelist[i], 'r')
      except IOError:
        logger.error('Cannot open: ' +  filelist[i])
        return
      currlines[i] = filehandlers[i].readline().strip()
    while True:
      # Assuming logs won't have far future dates - 1 yr max since sometimes people have logs in near future dates
      future_time = str( datetime.datetime.utcnow() + datetime.timedelta(days=365))
      min_ts = future_time
      for i in range(len(currlines)):
        if currlines[i] == "":
          continue
        ts = currlines[i].split(',')[0]
        if ts < min_ts:
          min_ts = ts
      if min_ts == future_time:
        break
      outwords = []
      outwords.append(min_ts)
      for i in range(len(currlines)):
        if currlines[i] == "":
          outwords.append(filler)
        else:
          ts = currlines[i].split(',')[0]
          val = currlines[i].split(',')[1]
          if ts == min_ts:
            outwords.append(val)
            currlines[i] = filehandlers[i].readline().strip()
          else:
            outwords.append(filler)
      outf.write( ','.join(outwords) + '\n' )

def nway_plotting(crossplots, metrics, output_directory, resource_path, graphing_library):
  listlen = len(crossplots)
  if listlen == 0:
    return ''
  i = 0
  correlated_plots = []
  #GC.appstop,all GC.alloc,GC.alloc-rate GC.promo,GC.gen0t,GC.gen0sys
  while i < listlen:
    plot = crossplots[i]
    vals = plot.split(',')
    i += 1
    if not 'all' in vals:
      plot_data = []
      for val in vals:
        csv_file = get_default_csv(output_directory, val)
        plot_data.append(PlotData(input_csv=csv_file, csv_column=1, series_name=sanitize_string(val), y_label=sanitize_string(val), precision=None, graph_height=500, graph_width=1200, graph_type='line'))
      png_name = get_merged_plot_link_name(vals)
      graphed, div_file = Metric.graphing_modules[graphing_library].graph_data(plot_data, output_directory, resource_path, png_name)
      if graphed:
        correlated_plots.append(div_file)
    else:
      vals.remove('all')
      for metric in metrics:
        for csv in metric.csv_files:
          csv_filename = csv.split('/')[-1]
          metric_name = '.'.join(csv_filename.split('.')[0:-1])
          if metric_name in vals:
            continue
          new_val = []
          new_val.extend(vals)
          new_val.append(metric_name)
          new_val_str = ','.join(new_val)
          crossplots.append(new_val_str)
          listlen += 1
  return correlated_plots

def normalize_float_for_display(data_val):
  try:
    data_val = float(data_val)
  except ValueError:
    return data_val
  if data_val > 1:
    return '%.2f' % round(data_val, 2)
  else:
    return '%s' % float('%.2g' % data_val)

def calculate_stats(data_list, stats_to_calculate = ['mean', 'std'], percentiles_to_calculate = []):
  """
  Calculate statistics for given data.

  :param list data_list: List of floats
  :param list stats_to_calculate: List of strings with statistics to calculate. Supported stats are defined in constant stats_to_numpy_method_map
  :param list percentiles_to_calculate: List of floats that defined which percentiles to calculate.
  :return: tuple of dictionaries containing calculated statistics and percentiles
  """
  stats_to_numpy_method_map = {
      'mean' : numpy.mean,
      'avg' : numpy.mean,
      'std' : numpy.std,
      'standard_deviation' : numpy.std,
      'median' : numpy.median,
      'min' : numpy.amin,
      'max' : numpy.amax
      }
  calculated_stats = {}
  calculated_percentiles = {}
  for stat in stats_to_calculate:
    if stat in stats_to_numpy_method_map.keys():
      calculated_stats[stat] = stats_to_numpy_method_map[stat](data_list)
    else:
      logger.error("Unsupported stat : " + str(stat))
  for percentile in percentiles_to_calculate:
    if isinstance(percentile, float) or isinstance(percentile, int):
      calculated_percentiles[percentile] = numpy.percentile(data_list, percentile)
    else:
      logger.error("Unsupported percentile requested (should be int or float): " + str(percentile))
  return calculated_stats, calculated_percentiles

def is_valid_file(filename):
  """
  Check if the specifed file exists and is not empty

  :param filename: full path to the file that needs to be checked
  :return: Status, Message
  """
  if os.path.exists(filename):
    if not os.path.getsize(filename):
      logger.warning('%s : file is empty.', filename)
      return False
  else:
    logger.warning('%s : file does not exist.', filename)
    return False
  return True

def detect_timestamp_format(timestamp):
  """
  Given an input timestamp string, determine what format is it likely in.

  :param string timestamp: the timestamp string for which we need to determine format
  :return: best guess timestamp format
  """
  time_formats = {'epoch': re.compile(r'^[0-9]{10}$'), 'epoch_ms': re.compile(r'^[0-9]{13}$'),
                  '%Y-%m-%d %H:%M:%S': re.compile(r'^[0-9]{4}-[0-1][0-9]-[0-3][0-9] [0-2][0-9]:[0-5][0-9]:[0-5][0-9]$'),
                  '%Y-%m-%dT%H:%M:%S': re.compile(r'^[0-9]{4}-[0-1][0-9]-[0-3][0-9]T[0-2][0-9]:[0-5][0-9]:[0-5][0-9]$'),
                  '%Y-%m-%d_%H:%M:%S': re.compile(r'^[0-9]{4}-[0-1][0-9]-[0-3][0-9]_[0-2][0-9]:[0-5][0-9]:[0-5][0-9]$'),
                  '%Y-%m-%d %H:%M:%S.%f': re.compile(r'^[0-9]{4}-[0-1][0-9]-[0-3][0-9] [0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9]+$'),
                  '%Y-%m-%dT%H:%M:%S.%f': re.compile(r'^[0-9]{4}-[0-1][0-9]-[0-3][0-9]T[0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9]+$'),
                  '%Y-%m-%d_%H:%M:%S.%f': re.compile(r'^[0-9]{4}-[0-1][0-9]-[0-3][0-9]_[0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9]+$'),
                  '%Y%m%d %H:%M:%S': re.compile(r'^[0-9]{4}[0-1][0-9][0-3][0-9] [0-2][0-9]:[0-5][0-9]:[0-5][0-9]$'),
                  '%Y%m%dT%H:%M:%S': re.compile(r'^[0-9]{4}[0-1][0-9][0-3][0-9]T[0-2][0-9]:[0-5][0-9]:[0-5][0-9]$'),
                  '%Y%m%d_%H:%M:%S': re.compile(r'^[0-9]{4}[0-1][0-9][0-3][0-9]_[0-2][0-9]:[0-5][0-9]:[0-5][0-9]$'),
                  '%Y%m%d %H:%M:%S.%f': re.compile(r'^[0-9]{4}[0-1][0-9][0-3][0-9] [0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9]+$'),
                  '%Y%m%dT%H:%M:%S.%f': re.compile(r'^[0-9]{4}[0-1][0-9][0-3][0-9]T[0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9]+$'),
                  '%Y%m%d_%H:%M:%S.%f': re.compile(r'^[0-9]{4}[0-1][0-9][0-3][0-9]_[0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9]+$'),
                  '%H:%M:%S': re.compile(r'^[0-2][0-9]:[0-5][0-9]:[0-5][0-9]$'),
                  '%H:%M:%S.%f': re.compile(r'^[0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9]+$'),
                  '%Y-%m-%dT%H:%M:%S.%f%z': re.compile(r'^[0-9]{4}-[0-1][0-9]-[0-3][0-9]T[0-2][0-9]:[0-5][0-9]:[0-5][0-9].[0-9]+[+-][0-9]{4}$'),
  }
  for time_format in time_formats:
    if re.match(time_formats[time_format],timestamp):
      return time_format
  return 'unknown'

def get_standardized_timestamp(timestamp, ts_format):
  """
  Given a timestamp string, return a time stamp in the format YYYY-MM-DD HH:MM:SS.sss. If no date is present in
  timestamp then today's date will be added as a prefix
  """
  if not timestamp:
    return None
  if timestamp == 'now':
    timestamp = str(datetime.datetime.now())
  if not ts_format:
    ts_format = detect_timestamp_format(timestamp)
  try:
    if ts_format == '%Y-%m-%d %H:%M:%S.%f':
      return timestamp
    elif ts_format == 'unknown':
      logger.error('Unable to determine timestamp format for : %s', timestamp)
      return -1
    elif ts_format == 'epoch':
      ts = datetime.datetime.utcfromtimestamp(int(timestamp)).strftime('%Y-%m-%d %H:%M:%S.%f')
    elif ts_format == 'epoch_ms':
      ts = datetime.datetime.utcfromtimestamp(float(timestamp) / 1000).strftime('%Y-%m-%d %H:%M:%S.%f')
    elif ts_format in ('%H:%M:%S', '%H:%M:%S.%f'):
      date_today = str(datetime.date.today())
      ts = datetime.datetime.strptime(date_today + ' ' + timestamp,'%Y-%m-%d ' + ts_format).strftime('%Y-%m-%d %H:%M:%S.%f')
    else:
      ts = datetime.datetime.strptime(timestamp,ts_format).strftime('%Y-%m-%d %H:%M:%S.%f')
  except ValueError:
    return -1    
  return ts

def set_sla(obj, metric, sub_metric, rules):
  """
  Extract SLAs from a set of rules
  """
  if not hasattr(obj, 'sla_map'):
    return False
  rules_list = rules.split()
  for rule in rules_list:
    if '<' in rule:
      stat, threshold = rule.split('<')
      sla = SLA(metric, sub_metric, stat, threshold, 'lt')
    elif '>' in rule:
      stat, threshold = rule.split('>')
      sla = SLA(metric, sub_metric, stat, threshold, 'gt')
    else:
      if hasattr(obj, 'logger'):
        obj.logger.error('Unsupported SLA type defined : ' + rule)
      sla = None
    obj.sla_map[metric][sub_metric][stat] = sla
    if hasattr(obj, 'sla_list'):
      obj.sla_list.append(sla)  # TODO : remove this once report has grading done in the metric tables
  return True

def check_slas(metric):
  """
  Check if all SLAs pass
  :return: 0 (if all SLAs pass) or the number of SLAs failures
  """
  if not hasattr(metric, 'sla_map'):
    return
  for metric_label in metric.sla_map.keys():
    for sub_metric in metric.sla_map[metric_label].keys():
      for stat_name in metric.sla_map[metric_label][sub_metric].keys():
        sla = metric.sla_map[metric_label][sub_metric][stat_name]
        if stat_name[0] == 'p' and hasattr(metric, 'calculated_percentiles'):
          if sub_metric in metric.calculated_percentiles.keys():
            percentile_num = int(stat_name[1:])
            if isinstance(percentile_num, float) or isinstance(percentile_num, int):
              if percentile_num in metric.calculated_percentiles[sub_metric].keys():
                if not sla.check_sla_passed(metric.calculated_percentiles[sub_metric][percentile_num]):
                  logger.info("Failed SLA for " + sub_metric)
                  metric.status = CONSTANTS.SLA_FAILED
        if sub_metric in metric.calculated_stats.keys() and hasattr(metric, 'calculated_stats'):
          if stat_name in metric.calculated_stats[sub_metric].keys():
            if not sla.check_sla_passed(metric.calculated_stats[sub_metric][stat_name]):
              logger.info("Failed SLA for " + sub_metric)
              metric.status = CONSTANTS.SLA_FAILED
  # Save SLA results in a file
  if len(metric.sla_map.keys()) > 0 and hasattr(metric, 'get_sla_csv'):
    sla_csv_file = metric.get_sla_csv()
    with open(sla_csv_file, 'w') as FH:
      for metric_label in metric.sla_map.keys():
        for sub_metric in metric.sla_map[metric_label].keys():
          for stat, sla in metric.sla_map[metric_label][sub_metric].items():
            FH.write('%s\n' % (sla.get_csv_repr()))

def parse_and_plot_single_metrics(metric, graph_timezone, outdir_default, indir_default, graphing_library,
                                  skip_plots):
  metric.graph_timezone = graph_timezone
  if metric.outdir is None:
    metric.outdir = os.path.normpath(outdir_default)

  updated_infile_list = []
  for infile in metric.infile_list:
    # handling both cases of local file or http download.
    if not infile.startswith('http://') and not infile.startswith('https://'):
      updated_infile_list.append(os.path.join(indir_default, infile))
    else:
      updated_infile_list.append(infile)
  metric.infile_list = updated_infile_list

  if not metric.ignore:
    if metric.collect():
      if metric.parse():
        metric.calc()
        metric.calculate_stats()
        check_slas(metric)
        if not skip_plots:
          metric.graph(graphing_library)
      else:
        logger.error('Parsing failed for metric: '  + metric.label)
    else:
      logger.error('Fetch/Collect failed for metric: ' + metric.label)

def init_logging(logger, log_file, log_level):
  """
  Initialize the naarad logger.
  :param: logger: logger object to initialize
  :param: log_file: log file name
  :param: log_level: log level (debug, info, warn, error)
  """
  with open(log_file, 'w'):
    pass
  numeric_level = getattr(logging, log_level.upper(), None) if log_level else logging.INFO
  if not isinstance(numeric_level, int):
    raise ValueError('Invalid log level: %s' % log_level)
  logger.setLevel(logging.DEBUG)
  fh = logging.FileHandler(log_file)
  fh.setLevel(logging.DEBUG)
  ch = logging.StreamHandler()
  ch.setLevel(numeric_level)
  formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
  fh.setFormatter(formatter)
  ch.setFormatter(formatter)
  logger.addHandler(fh)
  logger.addHandler(ch)
  return CONSTANTS.OK

def get_argument_parser():
  """
  Initialize list of valid arguments accepted by Naarad CLI
  :return: arg_parser: argeparse.ArgumentParser object initialized with naarad CLI parameters
  """
  arg_parser = argparse.ArgumentParser()
  arg_parser.add_argument('-c', '--config', help="file with specifications for each metric and graphs")
  arg_parser.add_argument('--start', help="Start time in the format of HH:MM:SS or YYYY-mm-dd_HH:MM:SS")
  arg_parser.add_argument('--end', help="End time in the format of HH:MM:SS or YYYY-mm-dd_HH:MM:SS")
  arg_parser.add_argument('-i', '--input_dir', help="input directory used to construct full path name of the metric infile")
  arg_parser.add_argument('-o', '--output_dir', help="output directory where the plots and Report.html will be generated")
  arg_parser.add_argument('-V', '--variables', action="append", help="User defined variables (in form key=value) for substitution in the config file. Config should have the variable names in format %%(key)s")
  arg_parser.add_argument('-s', '--show_config', help="Print config associated with the provided template name", action="store_true")
  arg_parser.add_argument('-l', '--log', help="log level")
  arg_parser.add_argument('-d', '--diff', nargs=2, help="Specify the location of two naarad reports to diff separated by a space. Can be local or http(s) locations. The first report is used as a baseline.", metavar=("report-1", "report-2"))
  arg_parser.add_argument('-n', '--no_plots', help="Don't generate plot images. Useful when you only want SLA calculations. Note that on-demand charts can still be generated through client-charting.", action="store_true")
  arg_parser.add_argument('-e', '--exit_code', help="optional argument to enable exit_code for naarad", action="store_true")
  #TODO(Ritesh) : Print a list of all templates supported with descriptions
  #arg_parser.add_argument('-l', '--list_templates', help="List all template configs", action="store_true")
  return arg_parser

def get_variables(args):
  """
  Return a dictionary of variables specified at CLI
  :param: args: Command Line Arguments namespace
  """
  variables_dict = {}
  if args.variables:
    for var in args.variables:
      words = var.split('=')
      variables_dict[words[0]] = words[1]
  return variables_dict

def validate_arguments(args):
  """
  Validate that the necessary argument for normal or diff analysis are specified.
  :param: args: Command line arguments namespace
  """
  if args.diff:
    if not args.output_dir:
      logger.error('No Output location specified')
      print_usage()
      sys.exit(0)
  # elif not (args.config and args.output_dir):
  elif not args.output_dir:
    print_usage()
    sys.exit(0)

def print_usage():
  """
  Print naarad CLI usage message
  """
  print ("Usage: "
               "\n To generate a diff report      : naarad -d report1 report2 -o <output_location> -c <optional: config-file> -e <optional: turn on exit code>"
               "\n To generate an analysis report : naarad -i <input_location> -o <output_location> -c <optional: config_file> -e <optional: turn on exit code> -n <optional: disable plotting of images>")


def discover_by_name(input_directory, output_directory):
  """
  Auto discover metric types from the files that exist in input_directory and return a list of metrics
  :param: input_directory: The location to scan for log files
  :param: output_directory: The location for the report
  """
  metric_list = []
  log_files = os.listdir(input_directory)
  for log_file in log_files:
    if log_file in CONSTANTS.SUPPORTED_FILENAME_MAPPING.keys():
      metric_list.append(initialize_metric(CONSTANTS.SUPPORTED_FILENAME_MAPPING[log_file], [log_file], None, output_directory, CONSTANTS.RESOURCE_PATH, CONSTANTS.SUPPORTED_FILENAME_MAPPING[log_file], None, None, {}, None, {}))
    else:
      logger.warning('Unable to determine metric type for file: %s', log_file)
  return metric_list

def initialize_metric(section, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics, other_options):
  """
  Initialize appropriate metric based on type of metric.
  :param: section: config section name or auto discovered metric type
  :param: infile_list: list of input log files for the metric
  :param: hostname: hostname associated with the logs origin
  :param: output_directory: report location
  :param: resource_path: resource path for report
  :param: label: label for config section or auto discovered metric type
  :param: ts_start: start time for analysis
  :param: ts_end: end time for analysis
  :param: rule_strings: list of slas
  :param: important_sub_metrics: list of important sub metrics
  :param: other_options: kwargs
  :return: metric object
  """
  bin_path = os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),'bin'))
  metric = None
  metric_type = section.split('-')[0]
  if metric_type in metric_classes:
    if 'SAR' in metric_type:
      metric = metric_classes['SAR'](section, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics, **other_options)
    else:
      metric = metric_classes[metric_type](section, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics, **other_options)
  else:
    metric = Metric(section, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics, **other_options)
  metric.bin_path = bin_path
  return metric

def initialize_aggregate_metric(section, aggr_hosts, aggr_metrics, metrics, outdir_default, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics, other_options):
  """
  Initialize aggregate metric
  :param: section: config section name
  :param: aggr_hosts: list of hostnames to aggregate
  :param: aggr_metrics: list of metrics to aggregate
  :param: metrics: list of metric objects associated with the current naarad analysis
  :param: outdir_default: report location
  :param: resource_path: resource path for report
  :param: label: label for config section
  :param: ts_start: start time for analysis
  :param: ts_end: end time for analysis
  :param: rule_strings: list of slas
  :param: important_sub_metrics: list of important sub metrics
  :param: other_options: kwargs
  :return: metric object
  """
  bin_path = os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),'bin'))
  metric = None
  metric_type = section.split('-')[0]
  metric = aggregate_metric_classes[metric_type](section, aggr_hosts, aggr_metrics, metrics, outdir_default, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics, **other_options)
  metric.bin_path = bin_path
  return metric



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
import datetime
import logging
import os
import re
import sys
import threading
import json
import time
from datetime import date
from naarad.metrics.metric import Metric
import naarad.utils
import naarad.naarad_constants as CONSTANTS

logger = logging.getLogger('naarad.metrics.linkedin_android_rum_metric')

class LinkedInAndroidRumMetric(Metric):
  """ 
  Class for LinkedIn Android RUM logs, deriving from class Metric 
  Note that this is for LinkedIn only
  """
  clock_format = '%Y-%m-%d %H:%M:%S'
  val_types = ('launch_time', 'nus_update_time')


  def __init__ (self, metric_type, infile_list, hostname, outdir, resource_path, label, ts_start, ts_end, rule_strings,
                important_sub_metrics, **other_options):
    Metric.__init__(self, metric_type, infile_list, hostname, outdir, resource_path, label, ts_start, ts_end, rule_strings,
                    important_sub_metrics)
    self.sub_metrics = self.val_types
    if not self.important_sub_metrics:
      self.important_sub_metrics = CONSTANTS.important_sub_metrics_import['LINKEDINANDROIDRUM']
    self.sub_metric_description = {
      "launch_time" :"the time taken to launch the client application",
      "nus_update_time" :"the time taken to update NUS list after launch"
    }


  # get start time stamp, launch time duration, and nus update time duration 
  def get_times(self, native):
    """
    get start time stamp, launch time duration, and nus update time duration from JSON object native
    :param JSON OBJECT native
    :return: LONG event time stamp, LONG launch time, and LONG nus update time
    """
    start_time = 0
    end_time = 0
    launch_time = 0
    nus_update_time = 0

    for item in native:
      if item[CONSTANTS.LIA_TIMING_NAME] == CONSTANTS.LIA_APP_ON_CREATE and item[CONSTANTS.LIA_START] is not None:
        start_time = item[CONSTANTS.LIA_START][CONSTANTS.LIA_LONG]
      if item[CONSTANTS.LIA_TIMING_NAME] == CONSTANTS.LIA_NUS_UPDATE:
        if item[CONSTANTS.LIA_TIMING_VALUE] is not None:
          nus_update_time = item[CONSTANTS.LIA_TIMING_VALUE][CONSTANTS.LIA_LONG]
        if item[CONSTANTS.LIA_START] is not None:
          end_time = item[CONSTANTS.LIA_START][CONSTANTS.LIA_LONG]

    if start_time == 0 or end_time == 0:
      time_stamp = 0
      launch_time = 0
    else:
      time_stamp = start_time
      launch_time = end_time - start_time
    return (time_stamp, launch_time, nus_update_time)

    
  # parse Android RUM logs
  def parse(self):
    # check if outdir exists, if not, create it
    if not os.path.isdir(self.outdir):
      os.makedirs(self.outdir)
    if not os.path.isdir(self.resource_directory):
      os.makedirs(self.resource_directory)

    results = {}
    ts = None

    # set output csv
    launch_time_file = self.get_csv('launch_time')
    nus_update_time_file = self.get_csv('nus_update_time')
    for input_file in self.infile_list:
      # get Android RUM input data: for each line, generate (timestamp, launch_time, nus_update_time)
      with open(input_file, 'r') as inf:
        for line in inf:
          try:
            data = json.loads(line)
          except ValueError:
            logger.warn("Invalid JSON Object at line: %s", line)
          if data[CONSTANTS.LIA_NATIVE_TIMINGS] is not None:
            native = data[CONSTANTS.LIA_NATIVE_TIMINGS][CONSTANTS.LIA_ARRAY]
            time_stamp, launch_time, nus_update_time = self.get_times(native)
            if launch_time != 0 and nus_update_time != 0:
              results[time_stamp] = (str(launch_time), str(nus_update_time))
    # Writing launch time and nus update time stats
    with open(launch_time_file, 'w') as launchtimef:
      with open(nus_update_time_file, 'w') as nusupdatetimef:
        for ts in sorted(results.iterkeys()):
          launchtimef.write( naarad.utils.get_standardized_timestamp(ts, 'epoch_ms') + ',' + results[ts][0] + '\n' )
          nusupdatetimef.write( naarad.utils.get_standardized_timestamp(ts, 'epoch_ms') + ',' + results[ts][1] + '\n' )
    self.csv_files.append(launch_time_file)
    self.csv_files.append(nus_update_time_file)
    return True




# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0
 
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
import os
import random
import logging

logger = logging.getLogger('naarad.graphing.dygraphs')

def graph_csv(output_directory, resource_path, csv_file, plot_title, output_filename, y_label=None, precision=None, graph_height="600", graph_width="1500"):
  """ Single metric graphing function """
  if not os.path.getsize(csv_file):
    return False, ""
  y_label = y_label or plot_title
  div_id = str(random.random())
  div_string = "<div id=\"%s\" style=\"width:%spx; height:%spx;\"></div>" % (div_id, graph_width, graph_height)
  script_string = """<script type=\"text/javascript\">
        g2 = new Dygraph(
          document.getElementById(\"""" + div_id + """"),
            \"""" + resource_path + '/' + os.path.basename(csv_file) +  """",
            {
                        xValueFormatter: Dygraph.dateString_,
                        xValueParser: function(x) {
                                        var date_components = x.split(" ");
                                        var supported_format = date_components[0] + 'T' + date_components[1];
                                        if(date_components[1].indexOf(".") == -1)
                                        {
                                          supported_format += ".0";
                                        }
                                        return Date.parse(supported_format);
                                        },
                        xTicker: Dygraph.dateTicker,
                        xlabel: "Time",
                        ylabel: \"""" + y_label + """",
                        title: \"""" + plot_title + """",
                        labels: ["Time",\"""" + y_label + """"]
            }          // options
        );
        </script>"""

  with open(os.path.join(output_directory, output_filename + '.div'), 'w') as div_file:
    div_file.write(div_string + script_string)
  #Ritesh: TODO Also generate PNGs if someone needs them separately
  return True, os.path.join(output_directory, output_filename + '.div')

def graph_data(list_of_plots, output_directory, resource_path, output_filename):
  if len(list_of_plots) > 0:
    plot = list_of_plots[0]
    success, div_file = graph_csv(output_directory=output_directory, resource_path=resource_path, csv_file=plot.input_csv, plot_title=plot.graph_title, output_filename=output_filename, y_label=plot.y_label, precision=None, graph_height=plot.graph_height, graph_width=plot.graph_width)
    if len(list_of_plots) > 1:
      logger.warning('dygraph module currently does not support co-relation of multiple plots. Only plotting %s', plot.graph_title)
    return success, div_file
  else:
    return False, None

def graph_data_on_the_same_graph(list_of_plots, output_directory, resource_path, output_filename):
  """ 
  graph_data_on_the_same_graph: put a list of plots on the same graph: currently it supports CDF
  """
  logger.warning('graph_data_on_the_same_graph is currently not supported in dygraph')
  return False, None



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0
 
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
import nose  
from nose.plugins.attrib import attr
import os
import shutil
import sys
import time
import uuid

# add the path of ~/naarad/src;   the testing py is under ~/naarad/test 
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'src')))

import naarad.httpdownload

# the port of local http server
port_test = 8011

#the temporary directory for testing, will remove it after done. 
tmp_dir = ''

# the testing download file (will be hosted from local http server)
test_input_file = 'bin/naarad'

def setup():
  start_http_server()
  create_tmp_dir()

def teardown():
  kill_http_server()  
  delete_tmp_dir()  
  
def start_http_server():
  '''start a local http server for testing'''
  global port_test
  command = 'python -m SimpleHTTPServer %s &' % port_test
  os.system(command)
  time.sleep(1)

def kill_http_server():
  '''After testing, kill the local http server'''
  command = 'pkill -f SimpleHTTPServer'
  os.system(command)
  time.sleep(1)

def create_tmp_dir():
  '''create a unique tmp dir to hold the downloaded local files'''
  ''' if the tmp dir grenerated already exists, then simply return'''
  ''' the user simply try again to generate another unique tmp dir'''
  global tmp_dir
  tmp_dir = os.path.join('./tmp/', str(uuid.uuid4()))
  if not os.path.exists(tmp_dir):
    os.makedirs(tmp_dir)
  else:
    print "the path of %s already exists, please try again." % tmp_dir
    return
  
def delete_tmp_dir():
  '''delete the tmp directory'''
  global tmp_dir
  shutil.rmtree(tmp_dir)
  
@attr('local')
def test_list_of_urls_no_output():
  ''' list of abosulute urls with no output file name'''
  global tmp_dir
  url = "http://localhost:8011/bin/naarad"
  outdir = tmp_dir

  if os.path.exists(os.path.join(outdir, "naarad")):
    os.remove(os.path.join(outdir, "naarad"));  
  
  output_file = naarad.httpdownload.download_url_single(url, outdir)
  
  assert os.path.exists(output_file),  "File of %s does not exist! " % output_file  
  
  if os.path.exists(os.path.join(outdir, "naarad")):
    os.remove(os.path.join(outdir, "naarad"));   


@attr('local')
def test_list_of_urls_with_output():
  ''' list of abosulute urls with output file name given'''
  global tmp_dir
  
  url = "http://localhost:8011/bin/naarad"
  outfile = "naarad.tmp"
  outdir = tmp_dir 
  
  if os.path.exists(os.path.join(outdir, "1a.html")):
    os.remove(os.path.join(outdir, "1a.html"));  
      
  output_file = naarad.httpdownload.download_url_single(url, outdir, outfile)

  assert os.path.exists(output_file),  "File of %s does not exist! " % output_file  
       
  if os.path.exists(os.path.join(outdir, "1a.html")):
    os.remove(os.path.join(outdir, "1a.html"));  
      
@attr('local')
def test_regex_urls():
  '''a seeding url, and a regex expression of urls '''
  global tmp_dir
  seed_url = "http://localhost:8011/test/httpdownload.html"
  outdir = tmp_dir
  regex = ".*"

  output_files = []
  output_files = naarad.httpdownload.download_url_regex(seed_url, outdir, regex)
  
  print output_files
  print 'abc'
  output_file = os.path.join(outdir, 'test_httpdownload.pyc')
  assert os.path.exists(output_file),  "File of %s does not exist! " % output_file  
  output_file = os.path.join(outdir, 'test_httpdownload.py')
  assert os.path.exists(output_file),  "File of %s does not exist! " % output_file  




# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0
 
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""

import logging

logger = logging.getLogger('naarad.sla')

class SLA(object):

  supported_sla_types = ('lt', '<', 'gt', '>', 'eq', '=')

  def __init__(self, metric, sub_metric, stat_name, threshold, sla_type):
    if sla_type not in self.supported_sla_types:
      logger.error('Unsupported sla type passed : ' + sla_type)
      return None
    self.metric = metric
    self.sub_metric = sub_metric
    self.stat_name = stat_name
    self.sla_type = sla_type
    self.is_processed = False
    self.threshold = None
    self.display = None
    if '%' in threshold:
      self.threshold = float(threshold.translate(None, '%'))
      self.display = '%'
    else:
      self.threshold = float(threshold)
      self.display = ''
    self.sla_passed = None
    self.stat_value = None

  def __str__(self):
    return "{0} of {1}, threshold: {2}, sla_type: {3}, sla_passed: {4}, display: {5}".format(self.stat_name, self.sub_metric, self.threshold, self.sla_type, self.sla_passed, self.display)

  def get_csv_repr(self):
    return "{0},{1},{2},{3},{4},{5}".format(self.sub_metric, self.stat_name, self.threshold, self.sla_type, self.stat_value, self.sla_passed)

  def check_sla_passed(self, stat_value):
    if self.sla_type in ('lt', '<'):
      self.grade_lt(stat_value)
    elif self.sla_type in ('gt', '>'):
      self.grade_gt(stat_value)
    elif self.sla_type in ('eq', '='):
      self.grade_eq(stat_value)
    else:
      logger.error('sla type is unsupported')
    self.stat_value = stat_value
    return self.sla_passed

  def grade_lt(self, stat_value):
    self.is_processed = True
    if stat_value >= self.threshold:
      self.sla_passed = False
    else:
      self.sla_passed = True

  def grade_gt(self, stat_value):
    self.is_processed = True
    if stat_value <= self.threshold:
      self.sla_passed = False
    else:
      self.sla_passed = True
  
  def grade_eq(self, stat_value):
    self.is_processed = True
    if stat_value == self.threshold:
      self.sla_passed = True
    else:
      self.sla_passed = False



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
NAARAD_LOG = 'naarad.log'

# Metric Constants
DEFAULT_SUMMARY_STATS = ['mean', 'std', 'p50', 'p75', 'p90', 'p95', 'p99', 'min', 'max']
important_sub_metrics_import = {
    'GC': ('GCPause', 'used', 'cmsIM', 'cmsCM', 'gen0t', 'g1-pause-young', 'g1-pause-mixed', 'g1-pause-remark', 'g1-pause-cleanup'),
    'LINKEDINANDROIDRUM': ('launch_time', 'nus_update_time'),
    'SAR-cpuusage': ('%sys', '%usr'),
    'SAR-device': ('%util', 'await'),
    'JMETER': ('Overall_Summary.ResponseTime', 'Overall_Summary.DataThroughput', 'Overall_Summary.qps')
}
device_type_metrics = ('SAR-cpuusage', 'SAR-cpuhz', 'SAR-device', 'SAR-dev', 'SAR-edev')
# Status CODES
OK = 0
SLA_FAILED = 1
COLLECT_FAILED = 2
PARSE_FAILED = 3
GRAPH_FAILED = 4
REPORT_FAILED = 5
ERROR = 6
INVALID_CONFIG = 7

# Report Constants
RESOURCE_PATH = 'resources'
DEFAULT_REPORT_TITLE = 'naarad analysis report'
STYLESHEET_INCLUDES = ['bootstrap.min.css', 'inversion.css']
JAVASCRIPT_INCLUDES = ['jquery-1.7.1.min.js', 'dygraph-combined.js', 'bootstrap.js', 'sorttable.js', 'naarad.js']
PLOTS_CSV_LIST_FILE = 'list.txt'
CDF_PLOTS_CSV_LIST_FILE = 'cdf_list.txt'
STATS_CSV_LIST_FILE = 'stats.txt'
SUMMARY_REPORT_FILE = 'summary_report.html'
CLIENT_CHARTING_FILE = 'report.html'
DIFF_REPORT_FILE = 'diff_report.html'
METRIC_REPORT_SUFFIX = '_report.html'
TEMPLATE_HEADER = 'default_report_header.html'
TEMPLATE_FOOTER = 'default_report_footer.html'
TEMPLATE_SUMMARY_CONTENT = 'default_summary_content.html'
TEMPLATE_SUMMARY_PAGE = 'default_summary_page.html'
TEMPLATE_METRIC_PAGE = 'default_metric_page.html'
TEMPLATE_CLIENT_CHARTING = 'default_client_charting_page.html'
TEMPLATE_DIFF_CLIENT_CHARTING = 'default_diff_client_charting_page.html'
TEMPLATE_DIFF_PAGE = 'default_diff_page.html'
SUBMETRIC_HEADER = 'sub_metric'

# Graphing constants
DEFAULT_GRAPHING_LIBRARY = 'matplotlib'

# Matplotlib Constants
COLOR_PALETTE = ['black', 'steelblue', 'm', 'red', 'cyan', 'g', 'orange', 'gray']
SUBPLOT_BOTTOM_OFFSET = 0.1
SUBPLOT_LEFT_OFFSET = 0.05
SUBPLOT_RIGHT_OFFSET = 0.95
SUBPLOT_TOP_OFFSET = 0
X_TICKS_FONTSIZE = 8
X_TICKS_DATEFORMAT = '%H:%M:%S'
Y_AXIS_OFFSET = 0.06
Y_LABEL_FONTSIZE = 10
Y_TICKS_FONTSIZE = 8
ZOOM_FACTOR = 0.02

# LinkedIn_Android_RUM Constants
LIA_TIMING_NAME = 'timingName'
LIA_TIMING_VALUE = 'timingValue'
LIA_START = 'start'
LIA_APP_ON_CREATE = 'linkedin_android_app_oncreate_time'
LIA_NUS_UPDATE = 'linkedin_android_nus_update_time'
LIA_LONG = 'long'
LIA_NATIVE_TIMINGS = 'nativeTimings'
LIA_ARRAY = 'array'


# Narrad Exit Code
SLA_FAILURE = 1

# RUN STEPS constants
PRE_ANALYSIS_RUN = 'pre'
DURING_ANALYSIS_RUN = 'in'
POST_ANALYSIS_RUN = 'post'
RUN_TYPE_WORKLOAD = 'workload'
SECONDS_TO_KILL_AFTER_SIGTERM = 5

#Auto Discover Metrics
SUPPORTED_FILENAME_MAPPING = {
  'gc.log' : 'GC',
  'perf-results.xml' : 'JMETER',
  'perf-result.xml' : 'JMETER',
  'proc.vmstat.out' : 'PROCVMSTAT',
  'procvmstat.out' : 'PROCVMSTAT',
  'proc.meminfo.out' : 'PROCMEMINFO',
  'procmeminfo.out' : 'PROCMEMINFO',
  'proc.zoneinfo.out' : 'PROCZONEINFO',
  'proczoneinfo.out' : 'PROCZONEINFO',
  'sar.cpuhz.out' : 'SAR-cpuhz',
  'sar.cpuusage.out' : 'SAR-cpuusage',
  'sar.device.out' : 'SAR-device',
  'sar.edev.out' : 'SAR-edev',
  'sar.etcp.out' : 'SAR-etcp',
  'sar.memory.out' : 'SAR-memory',
  'sar.memutil.out' : 'SAR-memutil',
  'sar.network.out' : 'SAR-network',
  'sar.paging.out' : 'SAR-paging',
  'sar.queue.out' : 'SAR-queue',
  'sar.sock.out' : 'SAR-sock',
  'sar.swapping.out' : 'SAR-swapping',
  'sar.switching.out' : 'SAR-switching',
  'sar.tcp.out' : 'SAR-tcp'
}




# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0
 
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
import datetime
import logging
import os

from naarad.metrics.metric import Metric
import naarad.utils
from naarad.naarad_constants import important_sub_metrics_import

logger = logging.getLogger('naarad.metrics.SARMetric')

class SARMetric(Metric):
  """ Class for SAR cpuusage logs, deriving from class Metric """
  supported_sar_types = ('SAR-cpuusage', 'SAR-cpuhz', 'SAR-device', 'SAR-memory', 'SAR-memutil', 'SAR-paging', 
      'SAR-etcp', 'SAR-tcp', 'SAR-dev', 'SAR-edev', 'SAR-sock', 'SAR-swapping')
  def __init__(self, metric_type, infile_list, hostname, outdir, resource_path, label, ts_start, ts_end, rule_strings,
               important_sub_metrics, **other_options):
    metric_type = self.extract_metric_name(metric_type)
    Metric.__init__(self, metric_type, infile_list,  hostname, outdir, resource_path, label, ts_start, ts_end, rule_strings,
                    important_sub_metrics)
    if not self.important_sub_metrics and self.metric_type in important_sub_metrics_import.keys():
      self.important_sub_metrics = important_sub_metrics_import[self.metric_type]
    self.options = None
    self.devices = None
    for (key, val) in other_options.iteritems():
      setattr(self, key, val.split())


  def extract_metric_name(self, metric_name):
    """
    Method to extract SAR metric names from the section given in the config. The SARMetric class assumes that 
    the section name will contain the SAR types listed in self.supported_sar_types tuple

    :param str metric_name: Section name from the config
    :return: str which identifies what kind of SAR metric the section represents
    """
    for metric_type in self.supported_sar_types:
      if metric_type in metric_name:
        return metric_type
    logger.error('Section [%s] does not contain a valid metric type, using type: "SAR-generic". Naarad works better '
                 'if it knows the metric type. Valid SAR metric names are: %s', metric_name, self.supported_sar_types)
    return 'SAR-generic'

  def get_csv(self, col, device=None):
    column = naarad.utils.sanitize_string(col)
    if device is None:
      outcsv = os.path.join(self.resource_directory, "{0}.{1}.csv".format(self.label, column))
      self.csv_column_map[outcsv] = col
    else:
      outcsv = os.path.join(self.resource_directory, "{0}.{1}.{2}.csv".format(self.label, device, column))
      self.csv_column_map[outcsv] = device + '.' + col
    return outcsv

  def parse(self):
  # Multiple day span not supported. Assumes time is between 0:00 AM to 11:59 PM, or 0:00 to 23:59
    if not os.path.isdir(self.outdir):
      os.makedirs(self.outdir)
    if not os.path.isdir(self.resource_directory):
      os.makedirs(self.resource_directory)
    data = {}
    for input_file in self.infile_list:
      with open(input_file, 'r') as infile:
        line = infile.readline()
        # Pre-processing
        try:
          datesar = line.split()[3].split('/')
          # year is not fully qualified - this will work till year 2999 :)
          if int(datesar[2]) < 1000:
            year = int(datesar[2]) + 2000
            datesar[2] = str(year)
        except IndexError:
          logger.error("Header not found for file: %s", input_file)
          logger.error("line: %s", line)
          return False
        date = datesar[2] + '-' + datesar[0] + '-' + datesar[1]
        infile.readline()   #skip blank line
        line = infile.readline()
        columns = line.split()
        if columns[1] in ('AM', 'PM'):
          ts_end_index = 2
        else:
          ts_end_index = 1
        if self.metric_type in self.device_types:
          columnstart = ts_end_index + 1
        else:
          columnstart = ts_end_index
        # Actually parsing data
        lines = infile.readlines()
        last_ts = None
        for i in range(len(lines)):
          # Skipping last line of the file since it could be malformed
          if i == len(lines) - 1:
            break
          line = lines[i]
          # Skipping header lines
          if 'Linux' in line or 'Average' in line or 'MHz' in line:
            continue
          words = line.split()
          if len(words) <= columnstart:
            continue
          ts = naarad.utils.convert_to_24hr_format( ' '.join(words[0:ts_end_index]) )
          if last_ts:
            if last_ts.startswith("23:") and ts.startswith("00:"):
              logger.info("Date rolling over")
              old_datetime = datetime.datetime.strptime(date, "%Y-%m-%d")
              new_datetime = old_datetime + datetime.timedelta(days=1)
              date = new_datetime.strftime("%Y-%m-%d")
          datetimestamp = date + ' ' + ts
          last_ts = ts
          if self.ts_out_of_range(datetimestamp):
            continue
          if self.metric_type in self.device_types:
            # Skipping headers that appear in the middle of the file
            if not naarad.utils.is_number( words[ts_end_index + 1] ):
              continue
            if self.devices and words[ts_end_index] not in self.devices:
              continue
            device = words[ts_end_index]
          else:
            # Skipping headers that appear in the middle of the file
            if not naarad.utils.is_number( words[ts_end_index] ):
              continue
            device = None
          datetimestamp = naarad.utils.reconcile_timezones(datetimestamp, self.timezone, self.graph_timezone)
          for i in range(columnstart, len(words)):
            if self.options and columns[i] not in self.options:
              continue
            outcsv = self.get_csv(columns[i], device)
            if outcsv in data:
              data[outcsv].append(datetimestamp + ',' + words[i])
            else:
              data[outcsv] = []
              data[outcsv].append( datetimestamp + ',' + words[i] )
    # Post processing, putting data in csv files
    for csv in data.keys():
      self.csv_files.append(csv)
      with open(csv, 'w') as csvf:
        csvf.write('\n'.join(sorted(data[csv])))
    return True




# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0
 
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""

import logging

logger = logging.getLogger('naarad.run_steps.run_step')

class Run_Step(object):
  """
  Base class that holds information about different kinds of "run steps" like Workload kickoff, Pre-run setup scripts,
  Post-run setup scripts
  """

  def __init__(self, run_type, run_cmd, call_type, run_order, run_rank, should_wait=True, kill_after_seconds=None):
    """
    Init method
    :param run_type: Type of run_step: "workload" only for now
    :param run_cmd: Details of command to be run. It could be a command or API call
    :param call_type: Kind of call -- local or remote
    :param run_order: When to run this w.r.t analysis. One of ('pre', 'in', 'post')
    :param run_rank: In what order to run this
    :param should_wait: Boolean whether naarad should wait for the run command to finish or not before moving on
    :param kill_after_seconds: Seconds for which the command should be run before being killed
    :return: None
    """
    self.run_type = run_type
    self.run_cmd = run_cmd
    self.call_type = call_type
    self.run_order = run_order
    self.run_rank = run_rank
    self.should_wait = should_wait
    self.kill_after_seconds = kill_after_seconds
    self.timer = None



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
from collections import defaultdict
import logging
from naarad.metrics.metric import Metric

logger = logging.getLogger('naarad.metrics.NetstatMetric')

class NetstatMetric(Metric):
  """
  Netstat metric
  The netstat output can be obtained by command of 'netstat -Ttp'
  'connections' is given by the user in config file. Each element contains two ends of the socket, they are keywords in the hostname and the port number;
  Note that port number can be 'ssh', and is optional
  'input_connections' is used internally, will contain a list of user interested connections;
  each element is a tuple of (host1,port1,host2,port2); (e.g., ('localhost','','','2344')
  'processes' is given by the user in config file;   It will contain user-input of process names (pid/process, or pid, or /process)
  'input_processes' is used internally, contains a list of pid/processes
  """
  connections = ''
  input_connections = []
  processes = ''
  input_processes = []
  
  def __init__ (self, metric_type, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end,
                rule_strings, important_sub_metrics, **other_options):
    Metric.__init__(self, metric_type, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end,
                    rule_strings, important_sub_metrics)
    self.sub_metrics = None
    for (key, val) in other_options.iteritems():
      setattr(self, key, val.split())

    self._extract_input_connections()
    self._extract_input_processes()

  def _get_tuple(self, fields):
    """
    :param fields: a list which contains either 0,1,or 2 values
    :return: a tuple with default values of '';
    """
    v1 = ''
    v2 = ''
    if len(fields) > 0:
      v1 = fields[0]
    if len(fields) > 1:
      v2 = fields[1]
    return v1, v2

  def _extract_input_connections(self):
    """
    Given user input of interested connections, it will extract the info and output a list of tuples.
    - input can be multiple values, separated by space;  
    - either host or port is optional
    - it may be just one end, 
    - e.g., "host1<->host2 host3<->  host1:port1<->host2"
    :return: None
    """
    for con in self.connections:
      ends = con.strip().split('<->')  # [host1:port1->host2]
      ends = filter(None, ends) #remove '' elements
      if len(ends) == 0:
        continue
      if len(ends) > 0: 
        host1, port1 = self._get_tuple(ends[0].split(':'))
      host2 = ''
      port2 = ''
      if len(ends) > 1: 
        host2, port2 = self._get_tuple(ends[1].split(':'))
      self.input_connections.append((host1,port1,host2,port2))

  def _extract_input_processes(self):
    """
    Given user input of interested processes, it will extract the info and output a list of tuples.
    - input can be multiple values, separated by space;
    - either pid or process_name is optional
    - e.g., "10001/python 10002/java cpp"
    :return: None
    """
    for proc in self.processes:
      ends = proc.split('/')
      pid, name = self._get_tuple(ends)
      self.input_processes.append((pid,name))

  def _match_host_port(self, host, port, cur_host, cur_port):
    """
    Determine whether user-specified (host,port) matches current (cur_host, cur_port)
    :param host,port: The user input of (host,port)
    :param cur_host, cur_port: The current connection
    :return: True or Not    
    """
    # if host is '', true;  if not '', it should prefix-match cur_host
    host_match = False
    if not host:
      host_match = True
    elif cur_host.startswith(host):  #allow for partial match
      host_match = True
      
    # if port is '', true;  if not '', it should exactly match cur_port
    port_match = False
    if not port:
      port_match = True
    elif port == cur_port:
      port_match = True

    return host_match and port_match

  def _match_processes(self, pid, name, cur_process):
    """
    Determine whether user-specified "pid/processes" contain this process
    :param pid: The user input of pid
    :param name: The user input of process name
    :param process: current process info
    :return: True or Not; (if both pid/process are given, then both of them need to match)
    """
    cur_pid, cur_name = self._get_tuple(cur_process.split('/'))

    pid_match = False
    if not pid:
      pid_match = True
    elif pid == cur_pid:
      pid_match = True

    name_match = False
    if not name:
      name_match = True
    elif name == cur_name:
      name_match = True

    return pid_match and name_match

  def _check_connection(self, local_end, remote_end, process):
    """ 
    Check whether the connection is of interest or not
    :param local_end: Local connection end point, e.g., 'host1:port1'
    :param remote_end: Remote connection end point, e.g., 'host2:port2'
    :param process: Current connection 's process info, e.g., '1234/firefox'
    :return: a tuple of (local_end, remote_end, True/False); e.g. ('host1_23232', 'host2_2222', True)
    """
    # check tcp end points
    cur_host1, cur_port1 = self._get_tuple(local_end.split(':'))
    cur_host2, cur_port2 = self._get_tuple(remote_end.split(':'))

    #check whether the connection is interested or not by checking user input
    host_port_is_interested = False
    for host1,port1,host2,port2 in self.input_connections:
      if self._match_host_port(host1, port1, cur_host1, cur_port1) and self._match_host_port(host2, port2, cur_host2, cur_port2):
        host_port_is_interested = True
        break
      if self._match_host_port(host1, port1, cur_host2, cur_port2) and self._match_host_port(host2, port2, cur_host1, cur_port1):
        host_port_is_interested = True
        break

    # check whether the connection is interested or not by checking process names given in the config
    process_is_interested = False
    for pid, name in self.input_processes:
      if self._match_processes(pid, name, process):
        process_is_interested = True
        break

    return cur_host1 + '_' + cur_port1, cur_host2 + '_' + cur_port2, host_port_is_interested and process_is_interested

  def _add_data_line(self, data, col, value, ts):
    """
    Append the data point to the dictionary of "data"
    :param data: The dictionary containing all data
    :param col: The sub-metric name e.g. 'host1_port1.host2_port2.SendQ'
    :param value: integer
    :param ts: timestamp
    :return: None
    """
    if col in self.column_csv_map:
      out_csv = self.column_csv_map[col]
    else:
      out_csv = self.get_csv(col)   #  column_csv_map[] is assigned in get_csv()
      data[out_csv] = []
    data[out_csv].append(ts + "," + value)

  def parse(self):
    """
    Parse the netstat output file
    :return: status of the metric parse
    """
    #sample netstat output: 2014-04-02 15:44:02.86612	tcp     9600      0 host1.localdomain.com.:21567 remote.remotedomain.com:51168 ESTABLISH pid/process
    data = {}  # stores the data of each sub-metric
    for infile in self.infile_list:
      logger.info('Processing : %s',infile)
      with open(infile) as fh:
        for line in fh:
          if 'ESTABLISHED' not in line:
            continue
          words = line.split()
          if len(words) < 8 or words[2] != 'tcp':
            continue
          ts = words[0] + " " + words[1]
          if self.ts_out_of_range(ts):
            continue
          # filtering based on user input; (local socket, remote socket, pid/process)
          local_end, remote_end, interested = self._check_connection(words[5], words[6], words[8])
          if interested:
            self._add_data_line(data, local_end + '.' + remote_end + '.RecvQ', words[3], ts)
            self._add_data_line(data, local_end + '.' + remote_end + '.SendQ', words[4], ts)
    #post processing, putting data in csv files;
    for csv in data.keys():
      self.csv_files.append(csv)
      with open(csv, 'w') as fh:
        fh.write('\n'.join(sorted(data[csv])))
    return True


# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at  http://www.apache.org/licenses/LICENSE-2.0
 
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
import numpy
import os
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from mpl_toolkits.axes_grid1 import host_subplot
import mpl_toolkits.axisartist as AA
import logging
import naarad.naarad_constants as CONSTANTS

logger = logging.getLogger('naarad.graphing.matplotlib')


def convert_to_mdate(date_str):
  try:
    mdate = mdates.strpdate2num('%Y-%m-%d %H:%M:%S.%f')(date_str)
  except:
    mdate = mdates.strpdate2num('%Y-%m-%d %H:%M:%S')(date_str)
  return mdate

# MPL-WA-07
# matplotlib does not rotate colors correctly when using multiple y axes. This method fills in that gap.
def get_current_color(index):
  return CONSTANTS.COLOR_PALETTE[index % len(CONSTANTS.COLOR_PALETTE)]


def get_graph_metadata(plots):
  height = 0
  width = 0
  title = ''
  for plot in plots:
    if plot.graph_height > height:
      height = plot.graph_height
    if plot.graph_width > width:
      width = plot.graph_width
    if title == '':
      title = plot.graph_title
    elif title != plot.graph_title:
      title = title + ',' + plot.graph_title
  return height/80, width/80, title


def curate_plot_list(plots):
  delete_nodes = []
  for plot in plots:
    if os.path.exists(plot.input_csv):
      if not os.path.getsize(plot.input_csv):
        logger.warning("%s file is empty. No plot corresponding to this file will be generated", plot.input_csv)
        delete_nodes.append(plot)
    else:
      logger.warning("%s file does not exist. No plot corresponding to this file will be generated", plot.input_csv)
      delete_nodes.append(plot)
  for node in delete_nodes:
    plots.remove(node)
  return plots


def graph_data(list_of_plots, output_directory, resource_path, output_filename):
  plots = curate_plot_list(list_of_plots)
  plot_count = len(plots)
  if plot_count == 0:
    return False, None
  graph_height, graph_width, graph_title = get_graph_metadata(list_of_plots)
  current_plot_count = 0
  fig, axis = plt.subplots()
  fig.set_size_inches(graph_width, graph_height)
  if plot_count < 2:
    fig.subplots_adjust(left=CONSTANTS.SUBPLOT_LEFT_OFFSET, bottom=CONSTANTS.SUBPLOT_BOTTOM_OFFSET, right=CONSTANTS.SUBPLOT_RIGHT_OFFSET)
  else:
    fig.subplots_adjust(left=CONSTANTS.SUBPLOT_LEFT_OFFSET, bottom=CONSTANTS.SUBPLOT_BOTTOM_OFFSET, right=CONSTANTS.SUBPLOT_RIGHT_OFFSET - CONSTANTS.Y_AXIS_OFFSET * (plot_count - 2))
  current_axis = axis
  for plot in plots:
    current_plot_count += 1
    logger.info('Processing: ' + plot.input_csv + ' [ ' + output_filename + ' ]')
    timestamp, yval = numpy.loadtxt(plot.input_csv, unpack=True, delimiter=',', converters={0: convert_to_mdate})
    maximum_yvalue = numpy.amax(yval) * (1.0 + CONSTANTS.ZOOM_FACTOR * current_plot_count)
    minimum_yvalue = numpy.amin(yval) * (1.0 - CONSTANTS.ZOOM_FACTOR * current_plot_count)

    if current_plot_count == 0:
      current_axis.yaxis.set_ticks_position('left')
    if current_plot_count > 1:
      current_axis = axis.twinx()
      current_axis.yaxis.grid(False)
      #Set right y-axis for additional plots
      current_axis.yaxis.set_ticks_position('right')
      #Offset the right y axis to avoid overlap
      current_axis.spines['right'].set_position(('axes', 1 + CONSTANTS.Y_AXIS_OFFSET * (current_plot_count-2)))
      current_axis.spines['right'].set_smart_bounds(False)
      current_axis.spines['right'].set_color(get_current_color(current_plot_count))
      current_axis.set_frame_on(True)
      current_axis.patch.set_visible(False)
    current_axis.set_ylabel(plot.y_label, color=get_current_color(current_plot_count), fontsize=CONSTANTS.Y_LABEL_FONTSIZE)
    current_axis.set_ylim([minimum_yvalue, maximum_yvalue])
    if plot.graph_type == 'line':
      current_axis.plot_date(x=timestamp, y=yval, linestyle='-', marker=None, color=get_current_color(current_plot_count))
    else:
      current_axis.plot_date(x=timestamp, y=yval, marker='.', color=get_current_color(current_plot_count))
    y_ticks = current_axis.get_yticklabels()
    for y_tick in y_ticks:
      y_tick.set_color(get_current_color(current_plot_count))
      y_tick.set_fontsize(CONSTANTS.Y_TICKS_FONTSIZE)
    for x_tick in current_axis.get_xticklabels():
      x_tick.set_fontsize(CONSTANTS.X_TICKS_FONTSIZE)
  axis.yaxis.grid(True)
  axis.xaxis.grid(True)
  axis.set_title(graph_title)
  axis.set_xlabel('Time')
  x_date_format = mdates.DateFormatter(CONSTANTS.X_TICKS_DATEFORMAT)
  axis.xaxis.set_major_formatter(x_date_format)
  plot_file_name = os.path.join(output_directory, output_filename + ".png")
  fig.savefig(plot_file_name)
  plt.close()
  #Create html fragment to be used for creation of the report
  with open(os.path.join(output_directory, output_filename + '.div'), 'w') as div_file:
    div_file.write('<div class="span12"><img src="' + resource_path + '/' + os.path.basename(plot_file_name) + '" id="' + os.path.basename(plot_file_name) + '" width="100%" height="auto"/></div><div class="span12"><p align=center>' + os.path.basename(plot_file_name) + '<br/></p></div>')
  return True, os.path.join(output_directory, output_filename + '.div')

def graph_data_on_the_same_graph(list_of_plots, output_directory, resource_path, output_filename):
  """ 
  graph_data_on_the_same_graph: put a list of plots on the same graph: currently it supports CDF
  """
  maximum_yvalue = -float('inf')
  minimum_yvalue = float('inf')
  plots = curate_plot_list(list_of_plots)
  plot_count = len(plots)
  if plot_count == 0:
    return False, None
  graph_height, graph_width, graph_title = get_graph_metadata(plots)
  current_plot_count = 0 
  fig, axis = plt.subplots()
  fig.set_size_inches(graph_width, graph_height)
  if plot_count < 2: 
    fig.subplots_adjust(left=CONSTANTS.SUBPLOT_LEFT_OFFSET, bottom=CONSTANTS.SUBPLOT_BOTTOM_OFFSET, right=CONSTANTS.SUBPLOT_RIGHT_OFFSET)
  else:
    fig.subplots_adjust(left=CONSTANTS.SUBPLOT_LEFT_OFFSET, bottom=CONSTANTS.SUBPLOT_BOTTOM_OFFSET, right=CONSTANTS.SUBPLOT_RIGHT_OFFSET - CONSTANTS.Y_AXIS_OFFSET * (plot_count - 2))
  # generate each plot on the graph
  for plot in plots:
    current_plot_count += 1
    logger.info('Processing: ' + plot.input_csv + ' [ ' + output_filename + ' ]')
    xval, yval = numpy.loadtxt(plot.input_csv, unpack=True, delimiter=',')
    axis.plot(xval, yval, linestyle='-', marker=None, color=get_current_color(current_plot_count), label=plot.plot_label)
    axis.legend()
    maximum_yvalue = max(maximum_yvalue, numpy.amax(yval) * (1.0 + CONSTANTS.ZOOM_FACTOR * current_plot_count))
    minimum_yvalue = min(minimum_yvalue, numpy.amin(yval) * (1.0 - CONSTANTS.ZOOM_FACTOR * current_plot_count))
  # set properties of the plots
  axis.yaxis.set_ticks_position('left')
  axis.set_xlabel(plots[0].x_label)
  axis.set_ylabel(plots[0].y_label, fontsize=CONSTANTS.Y_LABEL_FONTSIZE)
  axis.set_ylim([minimum_yvalue, maximum_yvalue])
  axis.yaxis.grid(True)
  axis.xaxis.grid(True)
  axis.set_title(graph_title)
  plot_file_name = os.path.join(output_directory, output_filename + ".png")
  fig.savefig(plot_file_name)
  plt.close()
  # Create html fragment to be used for creation of the report
  with open(os.path.join(output_directory, output_filename + '.div'), 'w') as div_file:
    div_file.write('<div class="span12"><img src="' + resource_path + '/' + os.path.basename(plot_file_name) + '" id="' + os.path.basename(plot_file_name) + '" width="100%" height="auto"/></div><div class="span12"><p align=center>' + os.path.basename(plot_file_name) + '<br/></p></div>')
  return True, os.path.join(output_directory, output_filename + '.div')



# coding=utf-8
"""
© 2013 LinkedIn Corp. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""

def import_modules(modules_dict, is_class_type=True):
  return_dict = {}
  for module_name, module_string in modules_dict.items():
    try:
      if is_class_type:
        file_name, class_name = module_string.rsplit('.', 1)
        mod = __import__(file_name, fromlist=[class_name])
        return_dict[module_name] = getattr(mod, class_name)
      else:
        return_dict[module_name] = __import__(module_string, fromlist=[module_string])
    except ImportError:
      pass
  return return_dict

metric_imports_dict = {
  'GC' : 'naarad.metrics.gc_metric.GCMetric',
  'INNOTOP' : 'naarad.metrics.innotop_metric.INNOMetric',
  'JMETER' : 'naarad.metrics.jmeter_metric.JmeterMetric',
  'LINKEDINANDROIDRUM' : 'naarad.metrics.linkedin_android_rum_metric.LinkedInAndroidRumMetric',
  'PROCVMSTAT' : 'naarad.metrics.procvmstat_metric.ProcVmstatMetric',
  'PROCMEMINFO' : 'naarad.metrics.procmeminfo_metric.ProcMeminfoMetric',
  'PROCZONEINFO' : 'naarad.metrics.proczoneinfo_metric.ProcZoneinfoMetric',
  'SAR' : 'naarad.metrics.sar_metric.SARMetric',
  'NETSTAT' : 'naarad.metrics.netstat_metric.NetstatMetric'
}

graphing_imports_dict = {
  'matplotlib':'naarad.graphing.matplotlib_naarad',
  'svg':'naarad.graphing.pygal_naarad'
}

aggregate_metric_imports_dict = {
  'CLUSTER': 'naarad.metrics.cluster_metric.ClusterMetric'
}

reporting_imports_dict = {
  'report' : 'naarad.reporting.report.Report'
}

metric_classes = import_modules(metric_imports_dict)

graphing_modules = import_modules(graphing_imports_dict, is_class_type=False)

aggregate_metric_classes = import_modules(aggregate_metric_imports_dict)

reporting_modules = import_modules(reporting_imports_dict)